from typing import Any, List, Type

from django.contrib.admin.options import InlineModelAdmin
from django.db.models.base import Model

class GenericInlineModelAdminChecks:
    def _check_exclude_of_parent_model(self, obj: GenericInlineModelAdmin, parent_model: Type[Model]) -> List[Any]: ...
    def _check_relation(self, obj: GenericInlineModelAdmin, parent_model: Type[Model]) -> List[Any]: ...

class GenericInlineModelAdmin(InlineModelAdmin):
    template: str = ...

class GenericStackedInline(GenericInlineModelAdmin): ...
class GenericTabularInline(GenericInlineModelAdmin): ...



from mypy.plugin import MethodContext
from mypy.types import AnyType, Instance
from mypy.types import Type as MypyType
from mypy.types import TypeOfAny

from mypy_django_plugin.django.context import DjangoContext
from mypy_django_plugin.lib import fullnames, helpers


def typecheck_queryset_filter(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    lookup_kwargs = ctx.arg_names[1]
    provided_lookup_types = ctx.arg_types[1]

    assert isinstance(ctx.type, Instance)

    if not ctx.type.args or not isinstance(ctx.type.args[0], Instance):
        return ctx.default_return_type

    model_cls_fullname = ctx.type.args[0].type.fullname
    model_cls = django_context.get_model_class_by_fullname(model_cls_fullname)
    if model_cls is None:
        return ctx.default_return_type

    for lookup_kwarg, provided_type in zip(lookup_kwargs, provided_lookup_types):
        if lookup_kwarg is None:
            continue
        if (isinstance(provided_type, Instance)
                and provided_type.type.has_base('django.db.models.expressions.Combinable')):
            provided_type = resolve_combinable_type(provided_type, django_context)

        lookup_type = django_context.resolve_lookup_expected_type(ctx, model_cls, lookup_kwarg)
        # Managers as provided_type is not supported yet
        if (isinstance(provided_type, Instance)
                and helpers.has_any_of_bases(provided_type.type, (fullnames.MANAGER_CLASS_FULLNAME,
                                                                  fullnames.QUERYSET_CLASS_FULLNAME))):
            return ctx.default_return_type

        helpers.check_types_compatible(ctx,
                                       expected_type=lookup_type,
                                       actual_type=provided_type,
                                       error_message=f'Incompatible type for lookup {lookup_kwarg!r}:')

    return ctx.default_return_type


def resolve_combinable_type(combinable_type: Instance, django_context: DjangoContext) -> MypyType:
    if combinable_type.type.fullname != fullnames.F_EXPRESSION_FULLNAME:
        # Combinables aside from F expressions are unsupported
        return AnyType(TypeOfAny.explicit)

    return django_context.resolve_f_expression_type(combinable_type)



from typing import Optional

from django.contrib.sessions.backends.base import SessionBase

class SessionStore(SessionBase):
    storage_path: str = ...
    file_prefix: str = ...
    def __init__(self, session_key: Optional[str] = ...) -> None: ...
    def clean(self) -> None: ...



from collections import OrderedDict
from typing import (
    TYPE_CHECKING, Any, Dict, Iterable, Iterator, List, Optional, Set, Union, cast,
)

from django.db.models.fields import Field
from django.db.models.fields.related import RelatedField
from django.db.models.fields.reverse_related import ForeignObjectRel
from mypy import checker
from mypy.checker import TypeChecker
from mypy.mro import calculate_mro
from mypy.nodes import (
    GDEF, MDEF, Block, ClassDef, Expression, MemberExpr, MypyFile, NameExpr, StrExpr, SymbolNode, SymbolTable,
    SymbolTableNode, TypeInfo, Var,
)
from mypy.plugin import (
    AttributeContext, CheckerPluginInterface, FunctionContext, MethodContext,
)
from mypy.types import AnyType, Instance, NoneTyp, TupleType
from mypy.types import Type as MypyType
from mypy.types import TypedDictType, TypeOfAny, UnionType

from mypy_django_plugin.lib import fullnames

if TYPE_CHECKING:
    from mypy_django_plugin.django.context import DjangoContext


def get_django_metadata(model_info: TypeInfo) -> Dict[str, Any]:
    return model_info.metadata.setdefault('django', {})


class IncompleteDefnException(Exception):
    pass


def lookup_fully_qualified_sym(fullname: str, all_modules: Dict[str, MypyFile]) -> Optional[SymbolTableNode]:
    if '.' not in fullname:
        return None
    module, cls_name = fullname.rsplit('.', 1)

    module_file = all_modules.get(module)
    if module_file is None:
        return None
    sym = module_file.names.get(cls_name)
    if sym is None:
        return None
    return sym


def lookup_fully_qualified_generic(name: str, all_modules: Dict[str, MypyFile]) -> Optional[SymbolNode]:
    sym = lookup_fully_qualified_sym(name, all_modules)
    if sym is None:
        return None
    return sym.node


def lookup_fully_qualified_typeinfo(api: TypeChecker, fullname: str) -> Optional[TypeInfo]:
    node = lookup_fully_qualified_generic(fullname, api.modules)
    if not isinstance(node, TypeInfo):
        return None
    return node


def lookup_class_typeinfo(api: TypeChecker, klass: type) -> Optional[TypeInfo]:
    fullname = get_class_fullname(klass)
    field_info = lookup_fully_qualified_typeinfo(api, fullname)
    return field_info


def reparametrize_instance(instance: Instance, new_args: List[MypyType]) -> Instance:
    return Instance(instance.type, args=new_args,
                    line=instance.line, column=instance.column)


def get_class_fullname(klass: type) -> str:
    return klass.__module__ + '.' + klass.__qualname__


def get_call_argument_by_name(ctx: Union[FunctionContext, MethodContext], name: str) -> Optional[Expression]:
    """
    Return the expression for the specific argument.
    This helper should only be used with non-star arguments.
    """
    if name not in ctx.callee_arg_names:
        return None
    idx = ctx.callee_arg_names.index(name)
    args = ctx.args[idx]
    if len(args) != 1:
        # Either an error or no value passed.
        return None
    return args[0]


def get_call_argument_type_by_name(ctx: Union[FunctionContext, MethodContext], name: str) -> Optional[MypyType]:
    """Return the type for the specific argument.

    This helper should only be used with non-star arguments.
    """
    if name not in ctx.callee_arg_names:
        return None
    idx = ctx.callee_arg_names.index(name)
    arg_types = ctx.arg_types[idx]
    if len(arg_types) != 1:
        # Either an error or no value passed.
        return None
    return arg_types[0]


def make_optional(typ: MypyType) -> MypyType:
    return UnionType.make_union([typ, NoneTyp()])


def parse_bool(expr: Expression) -> Optional[bool]:
    if isinstance(expr, NameExpr):
        if expr.fullname == 'builtins.True':
            return True
        if expr.fullname == 'builtins.False':
            return False
    return None


def has_any_of_bases(info: TypeInfo, bases: Iterable[str]) -> bool:
    for base_fullname in bases:
        if info.has_base(base_fullname):
            return True
    return False


def iter_bases(info: TypeInfo) -> Iterator[Instance]:
    for base in info.bases:
        yield base
        yield from iter_bases(base.type)


def get_private_descriptor_type(type_info: TypeInfo, private_field_name: str, is_nullable: bool) -> MypyType:
    """ Return declared type of type_info's private_field_name (used for private Field attributes)"""
    sym = type_info.get(private_field_name)
    if sym is None:
        return AnyType(TypeOfAny.explicit)

    node = sym.node
    if isinstance(node, Var):
        descriptor_type = node.type
        if descriptor_type is None:
            return AnyType(TypeOfAny.explicit)

        if is_nullable:
            descriptor_type = make_optional(descriptor_type)
        return descriptor_type
    return AnyType(TypeOfAny.explicit)


def get_field_lookup_exact_type(api: TypeChecker, field: Field) -> MypyType:
    if isinstance(field, (RelatedField, ForeignObjectRel)):
        lookup_type_class = field.related_model
        rel_model_info = lookup_class_typeinfo(api, lookup_type_class)
        if rel_model_info is None:
            return AnyType(TypeOfAny.from_error)
        return make_optional(Instance(rel_model_info, []))

    field_info = lookup_class_typeinfo(api, field.__class__)
    if field_info is None:
        return AnyType(TypeOfAny.explicit)
    return get_private_descriptor_type(field_info, '_pyi_lookup_exact_type',
                                       is_nullable=field.null)


def get_nested_meta_node_for_current_class(info: TypeInfo) -> Optional[TypeInfo]:
    metaclass_sym = info.names.get('Meta')
    if metaclass_sym is not None and isinstance(metaclass_sym.node, TypeInfo):
        return metaclass_sym.node
    return None


def add_new_class_for_module(module: MypyFile, name: str, bases: List[Instance],
                             fields: 'OrderedDict[str, MypyType]') -> TypeInfo:
    new_class_unique_name = checker.gen_unique_name(name, module.names)

    # make new class expression
    classdef = ClassDef(new_class_unique_name, Block([]))
    classdef.fullname = module.fullname + '.' + new_class_unique_name

    # make new TypeInfo
    new_typeinfo = TypeInfo(SymbolTable(), classdef, module.fullname)
    new_typeinfo.bases = bases
    calculate_mro(new_typeinfo)
    new_typeinfo.calculate_metaclass_type()

    # add fields
    for field_name, field_type in fields.items():
        var = Var(field_name, type=field_type)
        var.info = new_typeinfo
        var._fullname = new_typeinfo.fullname + '.' + field_name
        new_typeinfo.names[field_name] = SymbolTableNode(MDEF, var, plugin_generated=True)

    classdef.info = new_typeinfo
    module.names[new_class_unique_name] = SymbolTableNode(GDEF, new_typeinfo, plugin_generated=True)
    return new_typeinfo


def get_current_module(api: TypeChecker) -> MypyFile:
    current_module = None
    for item in reversed(api.scope.stack):
        if isinstance(item, MypyFile):
            current_module = item
            break
    assert current_module is not None
    return current_module


def make_oneoff_named_tuple(api: TypeChecker, name: str, fields: 'OrderedDict[str, MypyType]') -> TupleType:
    current_module = get_current_module(api)
    namedtuple_info = add_new_class_for_module(current_module, name,
                                               bases=[api.named_generic_type('typing.NamedTuple', [])],
                                               fields=fields)
    return TupleType(list(fields.values()), fallback=Instance(namedtuple_info, []))


def make_tuple(api: 'TypeChecker', fields: List[MypyType]) -> TupleType:
    # fallback for tuples is any builtins.tuple instance
    fallback = api.named_generic_type('builtins.tuple',
                                      [AnyType(TypeOfAny.special_form)])
    return TupleType(fields, fallback=fallback)


def convert_any_to_type(typ: MypyType, referred_to_type: MypyType) -> MypyType:
    if isinstance(typ, UnionType):
        converted_items = []
        for item in typ.items:
            converted_items.append(convert_any_to_type(item, referred_to_type))
        return UnionType.make_union(converted_items,
                                    line=typ.line, column=typ.column)
    if isinstance(typ, Instance):
        args = []
        for default_arg in typ.args:
            if isinstance(default_arg, AnyType):
                args.append(referred_to_type)
            else:
                args.append(default_arg)
        return reparametrize_instance(typ, args)

    if isinstance(typ, AnyType):
        return referred_to_type

    return typ


def make_typeddict(api: CheckerPluginInterface, fields: 'OrderedDict[str, MypyType]',
                   required_keys: Set[str]) -> TypedDictType:
    object_type = api.named_generic_type('mypy_extensions._TypedDict', [])
    typed_dict_type = TypedDictType(fields, required_keys=required_keys, fallback=object_type)
    return typed_dict_type


def resolve_string_attribute_value(attr_expr: Expression, ctx: Union[FunctionContext, MethodContext],
                                   django_context: 'DjangoContext') -> Optional[str]:
    if isinstance(attr_expr, StrExpr):
        return attr_expr.value

    # support extracting from settings, in general case it's unresolvable yet
    if isinstance(attr_expr, MemberExpr):
        member_name = attr_expr.name
        if isinstance(attr_expr.expr, NameExpr) and attr_expr.expr.fullname == 'django.conf.settings':
            if hasattr(django_context.settings, member_name):
                return getattr(django_context.settings, member_name)

    ctx.api.fail(f'Expression of type {type(attr_expr).__name__!r} is not supported', ctx.context)
    return None


def get_typechecker_api(ctx: Union[AttributeContext, MethodContext, FunctionContext]) -> TypeChecker:
    if not isinstance(ctx.api, TypeChecker):
        raise ValueError('Not a TypeChecker')
    return cast(TypeChecker, ctx.api)


def is_model_subclass_info(info: TypeInfo, django_context: 'DjangoContext') -> bool:
    return (info.fullname in django_context.all_registered_model_class_fullnames
            or info.has_base(fullnames.MODEL_CLASS_FULLNAME))


def check_types_compatible(ctx: Union[FunctionContext, MethodContext],
                           *, expected_type: MypyType, actual_type: MypyType, error_message: str) -> None:
    api = get_typechecker_api(ctx)
    api.check_subtype(actual_type, expected_type,
                      ctx.context, error_message,
                      'got', 'expected')


def add_new_sym_for_info(info: TypeInfo, *, name: str, sym_type: MypyType) -> None:
    # type=: type of the variable itself
    var = Var(name=name, type=sym_type)
    # var.info: type of the object variable is bound to
    var.info = info
    var._fullname = info.fullname + '.' + name
    var.is_initialized_in_class = True
    var.is_inferred = True
    info.names[name] = SymbolTableNode(MDEF, var,
                                       plugin_generated=True)



from typing import Any, Callable, Dict, List, Optional, Protocol, Sequence, Type, TypeVar, Union

from django.db.models.base import Model
from django.http.response import (
    HttpResponse as HttpResponse,
    HttpResponseRedirect as HttpResponseRedirect,
    HttpResponsePermanentRedirect as HttpResponsePermanentRedirect,
)

from django.db.models import Manager, QuerySet
from django.http import HttpRequest

def render_to_response(
    template_name: Union[str, Sequence[str]],
    context: Optional[Dict[str, Any]] = ...,
    content_type: Optional[str] = ...,
    status: Optional[int] = ...,
    using: Optional[str] = ...,
) -> HttpResponse: ...
def render(
    request: HttpRequest,
    template_name: Union[str, Sequence[str]],
    context: Optional[Dict[str, Any]] = ...,
    content_type: Optional[str] = ...,
    status: Optional[int] = ...,
    using: Optional[str] = ...,
) -> HttpResponse: ...

class SupportsGetAbsoluteUrl(Protocol): ...

def redirect(
    to: Union[Callable, str, SupportsGetAbsoluteUrl], *args: Any, permanent: bool = ..., **kwargs: Any
) -> Union[HttpResponseRedirect, HttpResponsePermanentRedirect]: ...

_T = TypeVar("_T", bound=Model)

def get_object_or_404(klass: Union[Type[_T], Manager[_T], QuerySet[_T]], *args: Any, **kwargs: Any) -> _T: ...
def get_list_or_404(klass: Union[Type[_T], Manager[_T], QuerySet[_T]], *args: Any, **kwargs: Any) -> List[_T]: ...
def resolve_url(to: Union[Callable, Model, str], *args: Any, **kwargs: Any) -> str: ...



from .base import Model as Model

from .aggregates import (
    Aggregate as Aggregate,
    Avg as Avg,
    Count as Count,
    Max as Max,
    Min as Min,
    StdDev as StdDev,
    Sum as Sum,
    Variance as Variance,
)

from .fields import (
    FieldDoesNotExist as FieldDoesNotExist,
    AutoField as AutoField,
    IntegerField as IntegerField,
    PositiveIntegerField as PositiveIntegerField,
    PositiveSmallIntegerField as PositiveSmallIntegerField,
    SmallIntegerField as SmallIntegerField,
    BigIntegerField as BigIntegerField,
    FloatField as FloatField,
    CharField as CharField,
    EmailField as EmailField,
    URLField as URLField,
    Field as Field,
    SlugField as SlugField,
    TextField as TextField,
    BooleanField as BooleanField,
    NullBooleanField as NullBooleanField,
    DateField as DateField,
    TimeField as TimeField,
    DateTimeField as DateTimeField,
    IPAddressField as IPAddressField,
    GenericIPAddressField as GenericIPAddressField,
    UUIDField as UUIDField,
    DecimalField as DecimalField,
    FilePathField as FilePathField,
    BinaryField as BinaryField,
    DurationField as DurationField,
    BigAutoField as BigAutoField,
    CommaSeparatedIntegerField as CommaSeparatedIntegerField,
    NOT_PROVIDED as NOT_PROVIDED,
)

from .fields.related import (
    ForeignKey as ForeignKey,
    OneToOneField as OneToOneField,
    ManyToManyField as ManyToManyField,
    ForeignObject as ForeignObject,
    ManyToManyRel as ManyToManyRel,
    ManyToOneRel as ManyToOneRel,
    OneToOneRel as OneToOneRel,
    ForeignObjectRel as ForeignObjectRel,
)
from .fields.files import (
    ImageField as ImageField,
    FileField as FileField,
    FieldFile as FieldFile,
    FileDescriptor as FileDescriptor,
)
from .fields.proxy import OrderWrt as OrderWrt

from .deletion import (
    CASCADE as CASCADE,
    SET_DEFAULT as SET_DEFAULT,
    SET_NULL as SET_NULL,
    DO_NOTHING as DO_NOTHING,
    PROTECT as PROTECT,
    SET as SET,
)

from .query import (
    Prefetch as Prefetch,
    QuerySet as QuerySet,
    RawQuerySet as RawQuerySet,
    prefetch_related_objects as prefetch_related_objects,
)

from .query_utils import Q as Q, FilteredRelation as FilteredRelation

from .lookups import Lookup as Lookup, Transform as Transform

from .expressions import (
    F as F,
    Expression as Expression,
    Subquery as Subquery,
    Exists as Exists,
    OrderBy as OrderBy,
    OuterRef as OuterRef,
    Case as Case,
    When as When,
    RawSQL as RawSQL,
    Value as Value,
    Func as Func,
    ExpressionWrapper as ExpressionWrapper,
    Combinable as Combinable,
    Col as Col,
    CombinedExpression as CombinedExpression,
    ExpressionList as ExpressionList,
    Random as Random,
    Ref as Ref,
    Window as Window,
    WindowFrame as WindowFrame,
    RowRange as RowRange,
    ValueRange as ValueRange,
)

from .manager import BaseManager as BaseManager, Manager as Manager

from . import lookups as lookups

from .aggregates import (
    Avg as Avg,
    Min as Min,
    Max as Max,
    Variance as Variance,
    StdDev as StdDev,
    Sum as Sum,
    Aggregate as Aggregate,
)

from .indexes import Index as Index

from . import signals as signals

from .constraints import (
    BaseConstraint as BaseConstraint,
    CheckConstraint as CheckConstraint,
    UniqueConstraint as UniqueConstraint,
)

from .enums import Choices as Choices, IntegerChoices as IntegerChoices, TextChoices as TextChoices



def supports_color() -> bool: ...

class Style:
    def DEBUG(self, text: str) -> str: ...
    def INFO(self, text: str) -> str: ...
    def SUCCESS(self, text: str) -> str: ...
    def WARNING(self, text: str) -> str: ...
    def ERROR(self, text: str) -> str: ...

def make_style(config_string: str = ...) -> Style: ...
def no_style() -> Style: ...
def color_style() -> Style: ...



from datetime import date
from io import BufferedReader, StringIO, TextIOWrapper
from typing import Any, Dict, Iterable, List, Mapping, Optional, Type, Union
from uuid import UUID

from django.core.management.base import OutputWrapper
from django.db.models.base import Model
from django.db.models.fields.related import ForeignKey, ManyToManyField

from django.db.models.fields import Field

class SerializerDoesNotExist(KeyError): ...
class SerializationError(Exception): ...

class DeserializationError(Exception):
    @classmethod
    def WithData(
        cls, original_exc: Exception, model: str, fk: Union[int, str], field_value: Optional[Union[List[str], str]]
    ) -> DeserializationError: ...

class M2MDeserializationError(Exception):
    original_exc: Exception = ...
    pk: List[str] = ...
    def __init__(self, original_exc: Exception, pk: Union[List[str], str]) -> None: ...

class ProgressBar:
    progress_width: int = ...
    output: None = ...
    total_count: int = ...
    prev_done: int = ...
    def __init__(self, output: Optional[Union[StringIO, OutputWrapper]], total_count: int) -> None: ...
    def update(self, count: int) -> None: ...

class Serializer:
    internal_use_only: bool = ...
    progress_class: Any = ...
    stream_class: Any = ...
    options: Any = ...
    stream: Any = ...
    selected_fields: Any = ...
    use_natural_foreign_keys: Any = ...
    use_natural_primary_keys: Any = ...
    first: bool = ...
    def serialize(
        self,
        queryset: Iterable[Model],
        *,
        stream: Optional[Any] = ...,
        fields: Optional[Any] = ...,
        use_natural_foreign_keys: bool = ...,
        use_natural_primary_keys: bool = ...,
        progress_output: Optional[Any] = ...,
        object_count: int = ...,
        **options: Any
    ) -> Any: ...
    def start_serialization(self) -> None: ...
    def end_serialization(self) -> None: ...
    def start_object(self, obj: Any) -> None: ...
    def end_object(self, obj: Any) -> None: ...
    def handle_field(self, obj: Any, field: Any) -> None: ...
    def handle_fk_field(self, obj: Any, field: Any) -> None: ...
    def handle_m2m_field(self, obj: Any, field: Any) -> None: ...
    def getvalue(self) -> Optional[Union[bytes, str]]: ...

class Deserializer:
    options: Any = ...
    stream: Any = ...
    def __init__(self, stream_or_string: Union[BufferedReader, TextIOWrapper, str], **options: Any) -> None: ...
    def __iter__(self) -> Deserializer: ...
    def __next__(self) -> None: ...

class DeserializedObject:
    object: Any = ...
    m2m_data: Dict[str, List[int]] = ...
    deferred_fields: Mapping[Field, Any]
    def __init__(
        self,
        obj: Model,
        m2m_data: Optional[Dict[str, List[int]]] = ...,
        deferred_fields: Optional[Mapping[Field, Any]] = ...,
    ) -> None: ...
    def save(self, save_m2m: bool = ..., using: Optional[str] = ..., **kwargs: Any) -> None: ...
    def save_deferred_fields(self, using: Optional[str] = ...) -> None: ...

def build_instance(Model: Type[Model], data: Dict[str, Optional[Union[date, int, str, UUID]]], db: str) -> Model: ...
def deserialize_m2m_values(field: ManyToManyField, field_value: Any, using: str) -> List[Any]: ...
def deserialize_fk_value(field: ForeignKey, field_value: Any, using: str) -> Any: ...



from collections import OrderedDict
from typing import List, Optional, Sequence, Type, Union

from django.core.exceptions import FieldError
from django.db.models.base import Model
from django.db.models.fields.related import RelatedField
from mypy.nodes import Expression, NameExpr
from mypy.plugin import FunctionContext, MethodContext
from mypy.types import AnyType, Instance
from mypy.types import Type as MypyType
from mypy.types import TypeOfAny

from mypy_django_plugin.django.context import (
    DjangoContext, LookupsAreUnsupported,
)
from mypy_django_plugin.lib import fullnames, helpers


def _extract_model_type_from_queryset(queryset_type: Instance) -> Optional[Instance]:
    for base_type in [queryset_type, *queryset_type.type.bases]:
        if (len(base_type.args)
                and isinstance(base_type.args[0], Instance)
                and base_type.args[0].type.has_base(fullnames.MODEL_CLASS_FULLNAME)):
            return base_type.args[0]
    return None


def determine_proper_manager_type(ctx: FunctionContext) -> MypyType:
    default_return_type = ctx.default_return_type
    assert isinstance(default_return_type, Instance)

    outer_model_info = helpers.get_typechecker_api(ctx).scope.active_class()
    if (outer_model_info is None
            or not outer_model_info.has_base(fullnames.MODEL_CLASS_FULLNAME)):
        return default_return_type

    return helpers.reparametrize_instance(default_return_type, [Instance(outer_model_info, [])])


def get_field_type_from_lookup(ctx: MethodContext, django_context: DjangoContext, model_cls: Type[Model],
                               *, method: str, lookup: str) -> Optional[MypyType]:
    try:
        lookup_field = django_context.resolve_lookup_into_field(model_cls, lookup)
    except FieldError as exc:
        ctx.api.fail(exc.args[0], ctx.context)
        return None
    except LookupsAreUnsupported:
        return AnyType(TypeOfAny.explicit)

    if isinstance(lookup_field, RelatedField) and lookup_field.column == lookup:
        related_model_cls = django_context.get_field_related_model_cls(lookup_field)
        if related_model_cls is None:
            return AnyType(TypeOfAny.from_error)
        lookup_field = django_context.get_primary_key_field(related_model_cls)

    field_get_type = django_context.get_field_get_type(helpers.get_typechecker_api(ctx),
                                                       lookup_field, method=method)
    return field_get_type


def get_values_list_row_type(ctx: MethodContext, django_context: DjangoContext, model_cls: Type[Model],
                             flat: bool, named: bool) -> MypyType:
    field_lookups = resolve_field_lookups(ctx.args[0], ctx, django_context)
    if field_lookups is None:
        return AnyType(TypeOfAny.from_error)

    typechecker_api = helpers.get_typechecker_api(ctx)
    if len(field_lookups) == 0:
        if flat:
            primary_key_field = django_context.get_primary_key_field(model_cls)
            lookup_type = get_field_type_from_lookup(ctx, django_context, model_cls,
                                                     lookup=primary_key_field.attname, method='values_list')
            assert lookup_type is not None
            return lookup_type
        elif named:
            column_types: 'OrderedDict[str, MypyType]' = OrderedDict()
            for field in django_context.get_model_fields(model_cls):
                column_type = django_context.get_field_get_type(typechecker_api, field,
                                                                method='values_list')
                column_types[field.attname] = column_type
            return helpers.make_oneoff_named_tuple(typechecker_api, 'Row', column_types)
        else:
            # flat=False, named=False, all fields
            field_lookups = []
            for field in django_context.get_model_fields(model_cls):
                field_lookups.append(field.attname)

    if len(field_lookups) > 1 and flat:
        typechecker_api.fail("'flat' is not valid when 'values_list' is called with more than one field", ctx.context)
        return AnyType(TypeOfAny.from_error)

    column_types = OrderedDict()
    for field_lookup in field_lookups:
        lookup_field_type = get_field_type_from_lookup(ctx, django_context, model_cls,
                                                       lookup=field_lookup, method='values_list')
        if lookup_field_type is None:
            return AnyType(TypeOfAny.from_error)
        column_types[field_lookup] = lookup_field_type

    if flat:
        assert len(column_types) == 1
        row_type = next(iter(column_types.values()))
    elif named:
        row_type = helpers.make_oneoff_named_tuple(typechecker_api, 'Row', column_types)
    else:
        row_type = helpers.make_tuple(typechecker_api, list(column_types.values()))

    return row_type


def extract_proper_type_queryset_values_list(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    # called on the Instance, returns QuerySet of something
    assert isinstance(ctx.type, Instance)
    assert isinstance(ctx.default_return_type, Instance)

    model_type = _extract_model_type_from_queryset(ctx.type)
    if model_type is None:
        return AnyType(TypeOfAny.from_omitted_generics)

    model_cls = django_context.get_model_class_by_fullname(model_type.type.fullname)
    if model_cls is None:
        return ctx.default_return_type

    flat_expr = helpers.get_call_argument_by_name(ctx, 'flat')
    if flat_expr is not None and isinstance(flat_expr, NameExpr):
        flat = helpers.parse_bool(flat_expr)
    else:
        flat = False

    named_expr = helpers.get_call_argument_by_name(ctx, 'named')
    if named_expr is not None and isinstance(named_expr, NameExpr):
        named = helpers.parse_bool(named_expr)
    else:
        named = False

    if flat and named:
        ctx.api.fail("'flat' and 'named' can't be used together", ctx.context)
        return helpers.reparametrize_instance(ctx.default_return_type, [model_type, AnyType(TypeOfAny.from_error)])

    # account for possible None
    flat = flat or False
    named = named or False

    row_type = get_values_list_row_type(ctx, django_context, model_cls,
                                        flat=flat, named=named)
    return helpers.reparametrize_instance(ctx.default_return_type, [model_type, row_type])


def resolve_field_lookups(lookup_exprs: Sequence[Expression], ctx: Union[FunctionContext, MethodContext],
                          django_context: DjangoContext) -> Optional[List[str]]:
    field_lookups = []
    for field_lookup_expr in lookup_exprs:
        field_lookup = helpers.resolve_string_attribute_value(field_lookup_expr, ctx, django_context)
        if field_lookup is None:
            return None
        field_lookups.append(field_lookup)
    return field_lookups


def extract_proper_type_queryset_values(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    # called on QuerySet, return QuerySet of something
    assert isinstance(ctx.type, Instance)
    assert isinstance(ctx.default_return_type, Instance)

    model_type = _extract_model_type_from_queryset(ctx.type)
    if model_type is None:
        return AnyType(TypeOfAny.from_omitted_generics)

    model_cls = django_context.get_model_class_by_fullname(model_type.type.fullname)
    if model_cls is None:
        return ctx.default_return_type

    field_lookups = resolve_field_lookups(ctx.args[0], ctx, django_context)
    if field_lookups is None:
        return AnyType(TypeOfAny.from_error)

    if len(field_lookups) == 0:
        for field in django_context.get_model_fields(model_cls):
            field_lookups.append(field.attname)

    column_types: 'OrderedDict[str, MypyType]' = OrderedDict()
    for field_lookup in field_lookups:
        field_lookup_type = get_field_type_from_lookup(ctx, django_context, model_cls,
                                                       lookup=field_lookup, method='values')
        if field_lookup_type is None:
            return helpers.reparametrize_instance(ctx.default_return_type, [model_type, AnyType(TypeOfAny.from_error)])

        column_types[field_lookup] = field_lookup_type

    row_type = helpers.make_typeddict(ctx.api, column_types, set(column_types.keys()))
    return helpers.reparametrize_instance(ctx.default_return_type, [model_type, row_type])



from typing import Any, List

from django.core.checks.messages import Warning

SECRET_KEY_MIN_LENGTH: int
SECRET_KEY_MIN_UNIQUE_CHARACTERS: int
W001: Any
W002: Any
W004: Any
W005: Any
W006: Any
W007: Any
W008: Any
W009: Any
W018: Any
W019: Any
W020: Any
W021: Any

def check_security_middleware(app_configs: None, **kwargs: Any) -> List[Warning]: ...
def check_xframe_options_middleware(app_configs: None, **kwargs: Any) -> List[Warning]: ...
def check_sts(app_configs: None, **kwargs: Any) -> List[Warning]: ...
def check_sts_include_subdomains(app_configs: None, **kwargs: Any) -> List[Warning]: ...
def check_sts_preload(app_configs: None, **kwargs: Any) -> List[Warning]: ...
def check_content_type_nosniff(app_configs: None, **kwargs: Any) -> List[Warning]: ...
def check_xss_filter(app_configs: None, **kwargs: Any) -> List[Warning]: ...
def check_ssl_redirect(app_configs: None, **kwargs: Any) -> List[Warning]: ...
def check_secret_key(app_configs: None, **kwargs: Any) -> List[Warning]: ...
def check_debug(app_configs: None, **kwargs: Any) -> List[Warning]: ...
def check_xframe_deny(app_configs: None, **kwargs: Any) -> List[Warning]: ...
def check_allowed_hosts(app_configs: None, **kwargs: Any) -> List[Warning]: ...



from typing import Any

from .utils import (
    DEFAULT_DB_ALIAS as DEFAULT_DB_ALIAS,
    DJANGO_VERSION_PICKLE_KEY as DJANGO_VERSION_PICKLE_KEY,
    ProgrammingError as ProgrammingError,
    IntegrityError as IntegrityError,
    OperationalError as OperationalError,
    DatabaseError as DatabaseError,
    DataError as DataError,
    NotSupportedError as NotSupportedError,
    InternalError as InternalError,
    InterfaceError as InterfaceError,
    ConnectionHandler as ConnectionHandler,
    Error as Error,
    ConnectionDoesNotExist as ConnectionDoesNotExist,
)

from . import migrations

connections: Any
router: Any
connection: Any

class DefaultConnectionProxy:
    def __getattr__(self, item: str) -> Any: ...
    def __setattr__(self, name: str, value: Any) -> None: ...
    def __delattr__(self, name: str) -> None: ...

def close_old_connections(**kwargs: Any) -> None: ...
def reset_queries(**kwargs: Any) -> None: ...



from pytest_mypy.collect import File
from pytest_mypy.item import YamlTestItem


def django_plugin_hook(test_item: YamlTestItem) -> None:
    custom_settings = test_item.parsed_test_data.get('custom_settings', '')
    installed_apps = test_item.parsed_test_data.get('installed_apps', None)

    if installed_apps and custom_settings:
        raise ValueError('"installed_apps" and "custom_settings" are not compatible, please use one or the other')

    if installed_apps is not None:
        # custom_settings is empty, add INSTALLED_APPS
        installed_apps += ['django.contrib.contenttypes']
        installed_apps_as_str = '(' + ','.join([repr(app) for app in installed_apps]) + ',)'
        custom_settings += 'INSTALLED_APPS = ' + installed_apps_as_str

    if 'SECRET_KEY' not in custom_settings:
        custom_settings = 'SECRET_KEY = "1"\n' + custom_settings

    django_settings_section = "\n[mypy.plugins.django-stubs]\n" \
                              "django_settings_module = mysettings"
    if not test_item.additional_mypy_config:
        test_item.additional_mypy_config = django_settings_section
    else:
        if '[mypy.plugins.django-stubs]' not in test_item.additional_mypy_config:
            test_item.additional_mypy_config += django_settings_section

    mysettings_file = File(path='mysettings.py', content=custom_settings)
    test_item.files.append(mysettings_file)



from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Type, Union

from django.db.models.base import Model

from .base import (
    DeserializationError as DeserializationError,
    DeserializedObject,
    Deserializer as Deserializer,
    M2MDeserializationError as M2MDeserializationError,
    SerializationError as SerializationError,
    Serializer as Serializer,
    SerializerDoesNotExist as SerializerDoesNotExist,
)

BUILTIN_SERIALIZERS: Any

class BadSerializer:
    internal_use_only: bool = ...
    exception: BaseException = ...
    def __init__(self, exception: BaseException) -> None: ...
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

def register_serializer(format: str, serializer_module: str, serializers: Optional[Dict[str, Any]] = ...) -> None: ...
def unregister_serializer(format: str) -> None: ...
def get_serializer(format: str) -> Union[Type[Serializer], BadSerializer]: ...
def get_serializer_formats() -> List[str]: ...
def get_public_serializer_formats() -> List[str]: ...
def get_deserializer(format: str) -> Union[Callable, Type[Deserializer]]: ...
def serialize(format: str, queryset: Iterable[Model], **options: Any) -> Any: ...
def deserialize(format: str, stream_or_string: Any, **options: Any) -> Iterator[DeserializedObject]: ...
def sort_dependencies(app_list: Iterable[Any]) -> List[Type[Model]]: ...



from typing import Optional

from mypy.plugin import ClassDefContext, MethodContext
from mypy.types import CallableType, Instance, NoneTyp
from mypy.types import Type as MypyType
from mypy.types import TypeType

from mypy_django_plugin.lib import helpers


def make_meta_nested_class_inherit_from_any(ctx: ClassDefContext) -> None:
    meta_node = helpers.get_nested_meta_node_for_current_class(ctx.cls.info)
    if meta_node is None:
        if not ctx.api.final_iteration:
            ctx.api.defer()
    else:
        meta_node.fallback_to_any = True


def get_specified_form_class(object_type: Instance) -> Optional[TypeType]:
    form_class_sym = object_type.type.get('form_class')
    if form_class_sym and isinstance(form_class_sym.type, CallableType):
        return TypeType(form_class_sym.type.ret_type)
    return None


def extract_proper_type_for_get_form(ctx: MethodContext) -> MypyType:
    object_type = ctx.type
    assert isinstance(object_type, Instance)

    form_class_type = helpers.get_call_argument_type_by_name(ctx, 'form_class')
    if form_class_type is None or isinstance(form_class_type, NoneTyp):
        form_class_type = get_specified_form_class(object_type)

    if isinstance(form_class_type, TypeType) and isinstance(form_class_type.item, Instance):
        return form_class_type.item

    if isinstance(form_class_type, CallableType) and isinstance(form_class_type.ret_type, Instance):
        return form_class_type.ret_type

    return ctx.default_return_type


def extract_proper_type_for_get_form_class(ctx: MethodContext) -> MypyType:
    object_type = ctx.type
    assert isinstance(object_type, Instance)

    form_class_type = get_specified_form_class(object_type)
    if form_class_type is None:
        return ctx.default_return_type

    return form_class_type



from sqlite3 import dbapi2 as Database
from typing import Any, Callable

from django.db.backends.base.base import BaseDatabaseWrapper

def decoder(conv_func: Callable) -> Callable: ...

class DatabaseWrapper(BaseDatabaseWrapper): ...

FORMAT_QMARK_REGEX: Any

class SQLiteCursorWrapper(Database.Cursor): ...



from django.core.exceptions import FieldDoesNotExist
from mypy.plugin import MethodContext
from mypy.types import AnyType, Instance
from mypy.types import Type as MypyType
from mypy.types import TypeOfAny

from mypy_django_plugin.django.context import DjangoContext
from mypy_django_plugin.lib import helpers


def _get_field_instance(ctx: MethodContext, field_fullname: str) -> MypyType:
    field_info = helpers.lookup_fully_qualified_typeinfo(helpers.get_typechecker_api(ctx),
                                                         field_fullname)
    if field_info is None:
        return AnyType(TypeOfAny.unannotated)
    return Instance(field_info, [AnyType(TypeOfAny.explicit), AnyType(TypeOfAny.explicit)])


def return_proper_field_type_from_get_field(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    # Options instance
    assert isinstance(ctx.type, Instance)

    # bail if list of generic params is empty
    if len(ctx.type.args) == 0:
        return ctx.default_return_type

    model_type = ctx.type.args[0]
    if not isinstance(model_type, Instance):
        return ctx.default_return_type

    model_cls = django_context.get_model_class_by_fullname(model_type.type.fullname)
    if model_cls is None:
        return ctx.default_return_type

    field_name_expr = helpers.get_call_argument_by_name(ctx, 'field_name')
    if field_name_expr is None:
        return ctx.default_return_type

    field_name = helpers.resolve_string_attribute_value(field_name_expr, ctx, django_context)
    if field_name is None:
        return ctx.default_return_type

    try:
        field = model_cls._meta.get_field(field_name)
    except FieldDoesNotExist as exc:
        # if model is abstract, do not raise exception, skip false positives
        if not model_cls._meta.abstract:
            ctx.api.fail(exc.args[0], ctx.context)
        return AnyType(TypeOfAny.from_error)

    field_fullname = helpers.get_class_fullname(field.__class__)
    return _get_field_instance(ctx, field_fullname)



from typing import Union

from django.http.request import HttpRequest
from django.http.response import HttpResponseRedirect

def shortcut(
    request: HttpRequest, content_type_id: Union[int, str], object_id: Union[int, str]
) -> HttpResponseRedirect: ...



from datetime import datetime
from decimal import Decimal
from re import RegexFlag
from typing import Any, Dict, List, Optional, Union, Pattern, Collection
from uuid import UUID

from django.core.files.base import File
from django.core.exceptions import ValidationError as ValidationError  # noqa: F401

EMPTY_VALUES: Any

_Regex = Union[str, Pattern[str]]

def _lazy_re_compile(regex: _Regex, flags: int = ...): ...

class RegexValidator:
    regex: _Regex = ...
    message: Any = ...
    code: str = ...
    inverse_match: bool = ...
    flags: int = ...
    def __init__(
        self,
        regex: Optional[_Regex] = ...,
        message: Optional[str] = ...,
        code: Optional[str] = ...,
        inverse_match: Optional[bool] = ...,
        flags: Optional[RegexFlag] = ...,
    ) -> None: ...
    def __call__(self, value: Optional[str]) -> None: ...

class URLValidator(RegexValidator):
    ul: str = ...
    ipv4_re: str = ...
    ipv6_re: str = ...
    hostname_re: Any = ...
    domain_re: Any = ...
    tld_re: Any = ...
    host_re: Any = ...
    schemes: Any = ...
    def __init__(self, schemes: Optional[Collection[str]] = ..., **kwargs: Any) -> None: ...

integer_validator: Any

def validate_integer(value: Optional[Union[float, str]]) -> None: ...

class EmailValidator:
    message: Any = ...
    code: str = ...
    user_regex: Any = ...
    domain_regex: Any = ...
    literal_regex: Any = ...
    domain_whitelist: Any = ...
    def __init__(
        self, message: Optional[str] = ..., code: Optional[str] = ..., whitelist: Optional[Collection[str]] = ...
    ) -> None: ...
    def __call__(self, value: Optional[str]) -> None: ...
    def validate_domain_part(self, domain_part: str) -> bool: ...

validate_email: Any
slug_re: Any
validate_slug: Any
slug_unicode_re: Any
validate_unicode_slug: Any

def validate_ipv4_address(value: str) -> None: ...
def validate_ipv6_address(value: str) -> None: ...
def validate_ipv46_address(value: str) -> None: ...

ip_address_validator_map: Any

def ip_address_validators(protocol: str, unpack_ipv4: bool) -> Any: ...
def int_list_validator(
    sep: str = ..., message: None = ..., code: str = ..., allow_negative: bool = ...
) -> RegexValidator: ...

validate_comma_separated_integer_list: Any

class BaseValidator:
    message: Any = ...
    code: str = ...
    limit_value: Any = ...
    def __init__(self, limit_value: Any, message: Optional[str] = ...) -> None: ...
    def __call__(self, value: Any) -> None: ...
    def compare(self, a: bool, b: bool) -> bool: ...
    def clean(self, x: Any) -> Any: ...

class MaxValueValidator(BaseValidator):
    message: Any = ...
    code: str = ...
    def compare(self, a: Union[datetime, Decimal, float], b: Union[datetime, Decimal, float]) -> bool: ...

class MinValueValidator(BaseValidator):
    message: Any = ...
    code: str = ...
    def compare(self, a: Union[datetime, Decimal, float], b: Union[datetime, Decimal, float]) -> bool: ...

class MinLengthValidator(BaseValidator):
    message: Any = ...
    code: str = ...
    def compare(self, a: int, b: int) -> bool: ...
    def clean(self, x: str) -> int: ...

class MaxLengthValidator(BaseValidator):
    message: Any = ...
    code: str = ...
    def compare(self, a: int, b: int) -> bool: ...
    def clean(self, x: Union[bytes, str]) -> int: ...

class DecimalValidator:
    messages: Any = ...
    max_digits: int = ...
    decimal_places: int = ...
    def __init__(self, max_digits: Optional[Union[int, str]], decimal_places: Optional[Union[int, str]]) -> None: ...
    def __call__(self, value: Decimal) -> None: ...

class FileExtensionValidator:
    message: Any = ...
    code: str = ...
    allowed_extensions: List[str] = ...
    def __init__(
        self,
        allowed_extensions: Optional[Collection[str]] = ...,
        message: Optional[str] = ...,
        code: Optional[str] = ...,
    ) -> None: ...
    def __call__(self, value: File) -> None: ...

def get_available_image_extensions() -> List[str]: ...
def validate_image_file_extension(value: File) -> None: ...

class ProhibitNullCharactersValidator:
    message: Any = ...
    code: str = ...
    def __init__(self, message: Optional[str] = ..., code: Optional[str] = ...) -> None: ...
    def __call__(self, value: Optional[Union[Dict[Any, Any], str, UUID]]) -> None: ...



from django.http.request import HttpRequest
from django.http.response import HttpResponse
from django.utils.deprecation import MiddlewareMixin

class XFrameOptionsMiddleware(MiddlewareMixin):
    def process_response(self, request: HttpRequest, response: HttpResponse) -> HttpResponse: ...
    def get_xframe_options_value(self, request: HttpRequest, response: HttpResponse) -> str: ...



from typing import Any, Dict, Optional, Union
from uuid import UUID

from django.db.models.options import Options
from django.template.context import RequestContext
from django.utils.safestring import SafeText

register: Any

def admin_urlname(value: Options, arg: SafeText) -> str: ...
def admin_urlquote(value: Union[int, str, UUID]) -> Union[int, str, UUID]: ...
def add_preserved_filters(
    context: Union[Dict[str, Union[Options, str]], RequestContext],
    url: str,
    popup: bool = ...,
    to_field: Optional[str] = ...,
) -> str: ...



import threading
from collections import OrderedDict
from typing import Any, Callable, DefaultDict, Dict, Iterable, List, Optional, Tuple, Type, Union

from django.db.migrations.state import AppConfigStub
from django.db.models.base import Model

from .config import AppConfig

class Apps:
    all_models: Dict[str, OrderedDict[str, Type[Model]]] = ...
    app_configs: OrderedDict[str, AppConfig] = ...
    stored_app_configs: List[Any] = ...
    apps_ready: bool = ...
    ready_event: threading.Event = ...
    loading: bool = ...
    _pending_operations: DefaultDict[Tuple[str, str], List]
    models_ready: bool = ...
    ready: bool = ...
    def __init__(self, installed_apps: Optional[Union[List[AppConfigStub], List[str], Tuple]] = ...) -> None: ...
    def populate(self, installed_apps: Union[List[AppConfigStub], List[str], Tuple] = ...) -> None: ...
    def check_apps_ready(self) -> None: ...
    def check_models_ready(self) -> None: ...
    def get_app_configs(self) -> Iterable[AppConfig]: ...
    def get_app_config(self, app_label: str) -> AppConfig: ...
    # it's not possible to support it in plugin properly now
    def get_models(self, include_auto_created: bool = ..., include_swapped: bool = ...) -> List[Type[Model]]: ...
    def get_model(self, app_label: str, model_name: Optional[str] = ..., require_ready: bool = ...) -> Type[Any]: ...
    def register_model(self, app_label: str, model: Type[Model]) -> None: ...
    def is_installed(self, app_name: str) -> bool: ...
    def get_containing_app_config(self, object_name: str) -> Optional[AppConfig]: ...
    def get_registered_model(self, app_label: str, model_name: str) -> Type[Model]: ...
    def get_swappable_settings_name(self, to_string: str) -> Optional[str]: ...
    def set_available_apps(self, available: List[str]) -> None: ...
    def unset_available_apps(self) -> None: ...
    def set_installed_apps(self, installed: Union[List[str], Tuple[str]]) -> None: ...
    def unset_installed_apps(self) -> None: ...
    def clear_cache(self) -> None: ...
    def lazy_model_operation(self, function: Callable, *model_keys: Any) -> None: ...
    def do_pending_operations(self, model: Type[Model]) -> None: ...

apps: Apps



from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union, Iterable

from django.contrib.auth.forms import AdminPasswordChangeForm
from django.forms.boundfield import BoundField
from django.forms.utils import ErrorDict
from django.forms.widgets import Media, Widget
from django.utils.safestring import SafeText

from django import forms
from django.db.models.fields import AutoField

ACTION_CHECKBOX_NAME: str

class ActionForm(forms.Form):
    action: Any = ...
    select_across: Any = ...

checkbox: Any

class AdminForm:
    prepopulated_fields: Any = ...
    model_admin: Any = ...
    readonly_fields: Any = ...
    def __init__(
        self,
        form: AdminPasswordChangeForm,
        fieldsets: List[Tuple[None, Dict[str, List[str]]]],
        prepopulated_fields: Dict[Any, Any],
        readonly_fields: Optional[Iterable[Any]] = ...,
        model_admin: Any = ...,
    ) -> None: ...
    def __iter__(self) -> Iterator[Fieldset]: ...
    @property
    def errors(self) -> ErrorDict: ...
    @property
    def non_field_errors(self) -> Callable: ...
    @property
    def media(self) -> Media: ...

class Fieldset:
    form: Any = ...
    classes: Any = ...
    description: Any = ...
    model_admin: Any = ...
    readonly_fields: Any = ...
    def __init__(
        self,
        form: Any,
        name: Optional[Any] = ...,
        readonly_fields: Optional[Iterable[Any]] = ...,
        fields: Any = ...,
        classes: Any = ...,
        description: Optional[Any] = ...,
        model_admin: Optional[Any] = ...,
    ) -> None: ...
    @property
    def media(self) -> Media: ...
    def __iter__(self) -> Iterator[Fieldline]: ...

class Fieldline:
    form: Any = ...
    fields: Any = ...
    has_visible_field: Any = ...
    model_admin: Any = ...
    readonly_fields: Any = ...
    def __init__(
        self, form: Any, field: Any, readonly_fields: Optional[Iterable[Any]] = ..., model_admin: Optional[Any] = ...
    ) -> None: ...
    def __iter__(self) -> Iterator[Union[AdminField, AdminReadonlyField]]: ...
    def errors(self) -> SafeText: ...

class AdminField:
    field: BoundField = ...
    is_first: bool = ...
    is_checkbox: bool = ...
    is_readonly: bool = ...
    def __init__(self, form: Any, field: Any, is_first: Any) -> None: ...
    def label_tag(self) -> SafeText: ...
    def errors(self) -> SafeText: ...

class AdminReadonlyField:
    field: Any = ...
    form: Any = ...
    model_admin: Any = ...
    is_first: Any = ...
    is_checkbox: bool = ...
    is_readonly: bool = ...
    empty_value_display: Any = ...
    def __init__(self, form: Any, field: Any, is_first: Any, model_admin: Optional[Any] = ...) -> None: ...
    def label_tag(self) -> SafeText: ...
    def contents(self) -> SafeText: ...

class InlineAdminFormSet:
    opts: Any = ...
    formset: Any = ...
    fieldsets: Any = ...
    model_admin: Any = ...
    readonly_fields: Any = ...
    prepopulated_fields: Any = ...
    classes: Any = ...
    has_add_permission: Any = ...
    has_change_permission: Any = ...
    has_delete_permission: Any = ...
    has_view_permission: Any = ...
    def __init__(
        self,
        inline: Any,
        formset: Any,
        fieldsets: Any,
        prepopulated_fields: Optional[Any] = ...,
        readonly_fields: Optional[Any] = ...,
        model_admin: Optional[Any] = ...,
        has_add_permission: bool = ...,
        has_change_permission: bool = ...,
        has_delete_permission: bool = ...,
        has_view_permission: bool = ...,
    ) -> None: ...
    def __iter__(self) -> Iterator[InlineAdminForm]: ...
    def fields(self) -> Iterator[Dict[str, Union[Dict[str, bool], bool, Widget, str]]]: ...
    def inline_formset_data(self) -> str: ...
    @property
    def forms(self): ...
    @property
    def non_form_errors(self) -> Callable: ...
    @property
    def media(self) -> Media: ...

class InlineAdminForm(AdminForm):
    formset: Any = ...
    original: Any = ...
    show_url: Any = ...
    absolute_url: Any = ...
    def __init__(
        self,
        formset: Any,
        form: Any,
        fieldsets: Any,
        prepopulated_fields: Any,
        original: Any,
        readonly_fields: Optional[Any] = ...,
        model_admin: Optional[Any] = ...,
        view_on_site_url: Optional[Any] = ...,
    ) -> None: ...
    def needs_explicit_pk_field(self) -> Union[bool, AutoField]: ...
    def pk_field(self) -> AdminField: ...
    def fk_field(self) -> AdminField: ...
    def deletion_field(self) -> AdminField: ...
    def ordering_field(self): ...

class InlineFieldset(Fieldset):
    formset: Any = ...
    def __init__(self, formset: Any, *args: Any, **kwargs: Any) -> None: ...

class AdminErrorList(forms.utils.ErrorList):
    def __init__(self, form: Any, inline_formsets: Any) -> None: ...



from typing import Callable, List, Optional, Set, Union

from django.contrib.auth import REDIRECT_FIELD_NAME as REDIRECT_FIELD_NAME  # noqa: F401

def user_passes_test(
    test_func: Callable, login_url: Optional[str] = ..., redirect_field_name: str = ...
) -> Callable: ...
def login_required(
    function: Optional[Callable] = ..., redirect_field_name: str = ..., login_url: Optional[str] = ...
) -> Callable: ...
def permission_required(
    perm: Union[List[str], Set[str], str], login_url: None = ..., raise_exception: bool = ...
) -> Callable: ...



from typing import Any, List

from django.core.checks.messages import CheckMessage

def check_user_model(app_configs: None = ..., **kwargs: Any) -> List[CheckMessage]: ...
def check_models_permissions(app_configs: None = ..., **kwargs: Any) -> List[Any]: ...



import itertools
import shutil
import subprocess
import sys
from argparse import ArgumentParser
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Pattern, Union

from git import Repo

from scripts.enabled_test_modules import (
    EXTERNAL_MODULES, IGNORED_ERRORS, IGNORED_MODULES, MOCK_OBJECTS,
)

DJANGO_COMMIT_REFS = {
    '2.2': 'e8b0903976077b951795938b260211214ed7fe41',
    '3.0': '7ec5962638144cbf4c2e47ea7d8dc02d1ce44394'
}
PROJECT_DIRECTORY = Path(__file__).parent.parent
DJANGO_SOURCE_DIRECTORY = PROJECT_DIRECTORY / 'django-sources'  # type: Path


def get_unused_ignores(ignored_message_freq: Dict[str, Dict[Union[str, Pattern], int]]) -> List[str]:
    unused_ignores = []
    for root_key, patterns in IGNORED_ERRORS.items():
        for pattern in patterns:
            if (ignored_message_freq[root_key][pattern] == 0
                    and pattern not in itertools.chain(EXTERNAL_MODULES, MOCK_OBJECTS)):
                unused_ignores.append(f'{root_key}: {pattern}')
    return unused_ignores


def is_pattern_fits(pattern: Union[Pattern, str], line: str):
    if isinstance(pattern, Pattern):
        if pattern.search(line):
            return True
    else:
        if pattern in line:
            return True
    return False


def is_ignored(line: str, test_folder_name: str, *, ignored_message_freqs: Dict[str, Dict[str, int]]) -> bool:
    if 'runtests' in line:
        return True

    if test_folder_name in IGNORED_MODULES:
        return True

    for pattern in IGNORED_ERRORS.get(test_folder_name, []):
        if is_pattern_fits(pattern, line):
            ignored_message_freqs[test_folder_name][pattern] += 1
            return True

    for pattern in IGNORED_ERRORS['__common__']:
        if is_pattern_fits(pattern, line):
            ignored_message_freqs['__common__'][pattern] += 1
            return True

    return False


def replace_with_clickable_location(error: str, abs_test_folder: Path) -> str:
    raw_path, _, error_line = error.partition(': ')
    fname, _, line_number = raw_path.partition(':')

    try:
        path = abs_test_folder.joinpath(fname).relative_to(PROJECT_DIRECTORY)
    except ValueError:
        # fail on travis, just show an error
        return error

    clickable_location = f'./{path}:{line_number or 1}'
    return error.replace(raw_path, clickable_location)


def get_django_repo_object() -> Repo:
    if not DJANGO_SOURCE_DIRECTORY.exists():
        DJANGO_SOURCE_DIRECTORY.mkdir(exist_ok=True, parents=False)
        return Repo.clone_from('https://github.com/django/django.git', DJANGO_SOURCE_DIRECTORY)
    else:
        repo = Repo(DJANGO_SOURCE_DIRECTORY)
        return repo


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('--django_version', choices=['2.2', '3.0'], required=True)
    args = parser.parse_args()

    commit_sha = DJANGO_COMMIT_REFS[args.django_version]
    repo = get_django_repo_object()
    if repo.head.commit.hexsha != commit_sha:
        repo.git.fetch('origin')
        repo.git.checkout(commit_sha)

    mypy_config_file = (PROJECT_DIRECTORY / 'scripts' / 'mypy.ini').absolute()
    mypy_cache_dir = Path(__file__).parent / '.mypy_cache'
    tests_root = DJANGO_SOURCE_DIRECTORY / 'tests'
    global_rc = 0

    try:
        mypy_options = ['--cache-dir', str(mypy_config_file.parent / '.mypy_cache'),
                        '--config-file', str(mypy_config_file),
                        '--show-traceback',
                        '--no-error-summary',
                        '--hide-error-context'
                        ]
        mypy_options += [str(tests_root)]

        import distutils.spawn

        mypy_executable = distutils.spawn.find_executable('mypy')
        mypy_argv = [mypy_executable, *mypy_options]
        completed = subprocess.run(
            mypy_argv,
            env={'PYTHONPATH': str(tests_root)},
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        output = completed.stdout.decode()

        ignored_message_freqs = defaultdict(lambda: defaultdict(int))

        sorted_lines = sorted(output.splitlines())
        for line in sorted_lines:
            try:
                path_to_error = line.split(':')[0]
                test_folder_name = path_to_error.split('/')[2]
            except IndexError:
                test_folder_name = 'unknown'

            if not is_ignored(line, test_folder_name,
                              ignored_message_freqs=ignored_message_freqs):
                global_rc = 1
                print(line)

        unused_ignores = get_unused_ignores(ignored_message_freqs)
        if unused_ignores:
            print('UNUSED IGNORES ------------------------------------------------')
            print('\n'.join(unused_ignores))

        sys.exit(global_rc)

    except BaseException as exc:
        shutil.rmtree(mypy_cache_dir, ignore_errors=True)
        raise exc



# Some errors occur for the test suite itself, and cannot be addressed via django-stubs. They should be ignored
# using this constant.
import re

IGNORED_MODULES = {'schema', 'gis_tests', 'admin_widgets', 'admin_filters', 'migrations',
                   'sitemaps_tests', 'staticfiles_tests', 'modeladmin', 'model_forms',
                   'generic_views', 'forms_tests', 'flatpages_tests', 'admin_utils',
                   'admin_ordering', 'admin_changelist', 'admin_views', 'redirects_tests',
                   'invalid_models_tests', 'i18n', 'migrate_signals', 'model_formsets',
                   'template_tests', 'template_backends', 'test_runner', 'admin_scripts',
                   'sites_tests', 'inline_formsets', 'foreign_object', 'cache', 'test_client', 'test_client_regress'}

MOCK_OBJECTS = ['MockRequest', 'MockCompiler', 'modelz', 'call_count', 'call_args_list',
                'call_args', 'MockUser', 'Xtemplate', 'DummyRequest', 'DummyUser', 'MinimalUser', 'DummyNode']
EXTERNAL_MODULES = ['psycopg2', 'PIL', 'selenium', 'oracle', 'mysql', 'sqlite3', 'sqlparse', 'tblib', 'numpy',
                    'bcrypt', 'argon2', 'xml.dom']
IGNORED_ERRORS = {
    '__common__': [
        *MOCK_OBJECTS,
        *EXTERNAL_MODULES,
        'Need type annotation for',
        'has no attribute "getvalue"',
        'Cannot assign to a method',
        'already defined',
        'Cannot assign to a type',
        '"HttpResponseBase" has no attribute',
        '"object" has no attribute',
        re.compile(r'"Callable\[.+, Any\]" has no attribute'),
        'has no attribute "deconstruct"',
        # private members
        re.compile(r'has no attribute ("|\')_[a-zA-Z_]+("|\')'),
        "'Settings' object has no attribute",
        '**Dict',
        re.compile(r"Expression of type '.*' is not supported"),
        'has incompatible type "object"',
        'undefined in superclass',
        'Argument after ** must be a mapping',
        'note:',
        re.compile(r'Item "None" of "[a-zA-Z_ ,\[\]]+" has no attribute'),
        '"Callable[..., None]" has no attribute',
        'does not return a value',
        'has no attribute "alternatives"',
        'gets multiple values for keyword argument',
        '"Handler" has no attribute',
        'Module has no attribute',
        'namedtuple',
        # TODO: see test in managers/test_managers.yml
        "Cannot determine type of",
        'cache_clear',
        'cache_info',
        'Incompatible types in assignment (expression has type "None", variable has type Module)',
        "Module 'django.contrib.messages.storage.fallback' has no attribute 'CookieStorage'",
        # TODO: not supported yet
        'GenericRelation',
        'RelatedObjectDoesNotExist',
        re.compile(r'"Field\[Any, Any\]" has no attribute '
                   r'"(through|field_name|field|get_related_field|related_model|related_name'
                   r'|get_accessor_name|empty_strings_allowed|many_to_many)"'),
        # TODO: multitable inheritance
        'ptr',
        'Incompatible types in assignment (expression has type "Callable[',
        'SimpleLazyObject',
        'Test',
        'Mixin" has no attribute',
        'Incompatible types in string interpolation',
        '"None" has no attribute',
        'has no attribute "assert',
    ],
    'aggregation': [
        re.compile(r'got "Optional\[(Author|Publisher)\]", expected "Union\[(Author|Publisher), Combinable\]"'),
        'Argument 2 for "super" not an instance of argument 1',
    ],
    'annotations': [
        'Incompatible type for "store" of "Employee" (got "Optional[Store]", expected "Union[Store, Combinable]")'
    ],
    'apps': [
        'Incompatible types in assignment (expression has type "str", target has type "type")',
    ],
    'auth_tests': [
        '"PasswordValidator" has no attribute "min_length"',
        'AbstractBaseUser',
        'Argument "password_validators" to "password_changed" has incompatible type "Tuple[Validator]"; '
        + 'expected "Optional[Sequence[PasswordValidator]]"',
        'Unsupported right operand type for in ("object")',
        'mock_getpass',
        'Unsupported left operand type for + ("Sequence[str]")',
        'AuthenticationFormWithInactiveUsersOkay',
        'Incompatible types in assignment (expression has type "Dict[str, Any]", variable has type "QueryDict")',
        'No overload variant of "int" matches argument type "AnonymousUser"',
        'expression has type "AnonymousUser", variable has type "User"',
    ],
    'basic': [
        'Unexpected keyword argument "unknown_kwarg" for "refresh_from_db" of "Model"',
        'Unexpected attribute "foo" for model "Article"',
        'has no attribute "touched"',
        'Incompatible types in assignment (expression has type "Type[CustomQuerySet]"',
        '"Manager[Article]" has no attribute "do_something"',
    ],
    'backends': [
        '"DatabaseError" has no attribute "pgcode"'
    ],
    'builtin_server': [
        '"ServerHandler" has no attribute',
    ],
    'bulk_create': [
        'has incompatible type "List[Country]"; expected "Iterable[TwoFields]"',
        'List item 1 has incompatible type "Country"; expected "ProxyCountry"',
    ],
    'check_framework': [
        'base class "Model" defined the type as "Callable',
        'Value of type "Collection[str]" is not indexable',
        'Unsupported target for indexed assignment',
    ],
    'constraints': [
        'Argument "condition" to "UniqueConstraint" has incompatible type "str"; expected "Optional[Q]"'
    ],
    'contenttypes_tests': [
        '"FooWithBrokenAbsoluteUrl" has no attribute "unknown_field"',
        'contenttypes_tests.models.Site',
        'Argument 1 to "set" of "RelatedManager" has incompatible type "SiteManager[Site]"',
    ],
    'custom_lookups': [
        'in base class "SQLFuncMixin"',
        'has no attribute "name"',
    ],
    'custom_columns': [
        "Cannot resolve keyword 'firstname' into field",
    ],
    'custom_pk': [
        '"Employee" has no attribute "id"',
    ],
    'custom_managers': [
        'Unsupported dynamic base class',
        'Incompatible types in assignment (expression has type "CharField',
        'Item "Book" of "Optional[Book]" has no attribute "favorite_avg"',
    ],
    'csrf_tests': [
        'expression has type "property", base class "HttpRequest" defined the type as "QueryDict"',
        'expression has type "Dict[<nothing>, <nothing>]", variable has type "SessionBase"',
    ],
    'dates': [
        'Too few arguments for "dates" of',
    ],
    'dbshell': [
        'Incompatible types in assignment (expression has type "None"',
    ],
    'defer': [
        'Too many arguments for "refresh_from_db" of "Model"'
    ],
    'delete': [
        'Incompatible type for lookup \'pk\': (got "Optional[int]", expected "Union[str, int]")',
    ],
    'dispatch': [
        'Item "str" of "Union[ValueError, str]" has no attribute "args"'
    ],
    'deprecation': [
        '"Manager" has no attribute "old"',
        '"Manager" has no attribute "new"'
    ],
    'db_functions': [
        '"FloatModel" has no attribute',
        'Incompatible types in assignment (expression has type "Optional[Any]", variable has type "FloatModel")'
    ],
    'decorators': [
        '"Type[object]" has no attribute "method"'
    ],
    'expressions_window': [
        'has incompatible type "str"'
    ],
    'file_uploads': [
        'has no attribute "content_type"',
    ],
    'file_storage': [
        'Incompatible types in assignment (expression has type "None", variable has type "str")',
        'Property "base_url" defined in "FileSystemStorage" is read-only',
    ],
    'files': [
        'Incompatible types in assignment (expression has type "IOBase", variable has type "File")',
        'Argument 1 to "write" of "SpooledTemporaryFile"',
    ],
    'filtered_relation': [
        'has no attribute "name"',
    ],
    'fixtures': [
        'Incompatible types in assignment (expression has type "int", target has type "Iterable[str]")',
        'Incompatible types in assignment (expression has type "SpyManager[Spy]"',
    ],
    'fixtures_regress': [
        'Unsupported left operand type for + ("None")',
    ],
    'from_db_value': [
        '"Cash" has no attribute',
        'Argument 1 to "__str__" of "Decimal"',
    ],
    'get_object_or_404': [
        'Argument 1 to "get_object_or_404" has incompatible type "str"; '
        + 'expected "Union[Type[<nothing>], QuerySet[<nothing>]]"',
        'Argument 1 to "get_list_or_404" has incompatible type "List[Type[Article]]"; '
        + 'expected "Union[Type[<nothing>], QuerySet[<nothing>]]"',
        'CustomClass',
    ],
    'generic_relations': [
        "Cannot resolve keyword 'vegetable' into field",
    ],
    'generic_relations_regress': [
        '"Link" has no attribute',
    ],
    'httpwrappers': [
        'Argument 2 to "appendlist" of "QueryDict"',
        'Incompatible types in assignment (expression has type "int", target has type "Union[str, List[str]]")',
        'Argument 1 to "fromkeys" of "QueryDict" has incompatible type "int"',
    ],
    'humanize_tests': [
        'Argument 1 to "append" of "list" has incompatible type "None"; expected "str"',
    ],
    'lookup': [
        'Unexpected keyword argument "headline__startswith" for "in_bulk" of',
        'is called with more than one field',
        "Cannot resolve keyword 'pub_date_year' into field",
        "Cannot resolve keyword 'blahblah' into field",
    ],
    'logging_tests': [
        re.compile(r'Argument [0-9] to "makeRecord" of "Logger"'),
        '"LogRecord" has no attribute "request"',
    ],
    'm2m_regress': [
        "Cannot resolve keyword 'porcupine' into field",
        'Argument 1 to "set" of "RelatedManager" has incompatible type "int"',
    ],
    'mail': [
        'List item 1 has incompatible type "None"; expected "str"',
        'Argument 1 to "push" of "SMTPChannel" has incompatible type "str"; expected "bytes"',
        'Value of type "Union[List[Message], str, bytes, None]" is not indexable',
        'Incompatible types in assignment '
        + '(expression has type "bool", variable has type "Union[SMTP_SSL, SMTP, None]")',
        re.compile(
            r'Item "(int|str)" of "Union\[Message, str, int, Any\]" has no attribute "(get_content_type|get_filename)"'
        )
    ],
    'messages_tests': [
        'List item 0 has incompatible type "Dict[str, Message]"; expected "Message"',
        'Too many arguments',
        'CustomRequest',
    ],
    'middleware': [
        re.compile(r'"(HttpRequest|WSGIRequest)" has no attribute'),
    ],
    'many_to_many': [
        '(expression has type "List[Article]", variable has type "RelatedManager[Article]"',
        '"add" of "RelatedManager" has incompatible type "Article"; expected "Union[Publication, int]"',
    ],
    'many_to_one': [
        'Incompatible type for "parent" of "Child" (got "None", expected "Union[Parent, Combinable]")',
        'Incompatible type for "parent" of "Child" (got "Child", expected "Union[Parent, Combinable]")',
        'expression has type "List[<nothing>]", variable has type "RelatedManager[Article]"',
        '"Reporter" has no attribute "cached_query"',
        'to "add" of "RelatedManager" has incompatible type "Reporter"; expected "Union[Article, int]"',
    ],
    'middleware_exceptions': [
        'Argument 1 to "append" of "list" has incompatible type "Tuple[Any, Any]"; expected "str"'
    ],
    'model_fields': [
        'Item "Field[Any, Any]" of "Union[Field[Any, Any], ForeignObjectRel]" has no attribute',
        'Incompatible types in assignment (expression has type "Type[Person',
        'Incompatible types in assignment (expression has type "FloatModel", variable has type',
        '"ImageFile" has no attribute "was_opened"',
        'Incompatible type for "size" of "FloatModel" (got "object", expected "Union[float, int, str, Combinable]")',
        'Incompatible type for "value" of "IntegerModel" (got "object", expected',
    ],
    'model_indexes': [
        'Argument "condition" to "Index" has incompatible type "str"; expected "Optional[Q]"'
    ],
    'model_inheritance': [
        'base class "AbstractBase" defined',
        'base class "AbstractModel" defined',
        'Definition of "name" in base class "ConcreteParent"',
        ' Definition of "name" in base class "AbstractParent"',
        'referent_references',
        "Cannot resolve keyword 'attached_comment_set' into field"
    ],
    'model_meta': [
        'List item 0 has incompatible type "str"; expected "Union[Field[Any, Any], ForeignObjectRel]"'
    ],
    'model_regress': [
        'Incompatible type for "department" of "Worker"',
        '"PickledModel" has no attribute',
        '"Department" has no attribute "evaluate"',
        'Unsupported target for indexed assignment',
    ],
    'model_formsets_regress': [
        'Incompatible types in assignment (expression has type "int", target has type "str")',
    ],
    'model_options': [
        'expression has type "Dict[str, Type[Model]]", target has type "OrderedDict',
    ],
    'model_enums': [
        "'bool' is not a valid base class",
    ],
    'multiple_database': [
        'Unexpected attribute "extra_arg" for model "Book"'
    ],
    'null_queries': [
        "Cannot resolve keyword 'foo' into field"
    ],
    'order_with_respect_to': [
        '"Dimension" has no attribute "set_component_order"',
    ],
    'one_to_one': [
        'expression has type "None", variable has type "UndergroundBar"',
        'Item "OneToOneField[Union[Place, Combinable], Place]" '
        + 'of "Union[OneToOneField[Union[Place, Combinable], Place], Any]"',
    ],
    'pagination': [
        '"int" not callable',
    ],
    'postgres_tests': [
        'DummyArrayField',
        'DummyJSONField',
        'Incompatible types in assignment (expression has type "Type[Field[Any, Any]]',
        'Argument "encoder" to "JSONField" has incompatible type "DjangoJSONEncoder";',
        '("None" and "SearchQuery")',
    ],
    'properties': [
        re.compile('Unexpected attribute "(full_name|full_name_2)" for model "Person"')
    ],
    'prefetch_related': [
        '"Person" has no attribute "houses_lst"',
        '"Book" has no attribute "first_authors"',
        '"Book" has no attribute "the_authors"',
        'Incompatible types in assignment (expression has type "List[Room]", variable has type "Manager[Room]")',
        'Item "Room" of "Optional[Room]" has no attribute "house_attr"',
        'Item "Room" of "Optional[Room]" has no attribute "main_room_of_attr"',
        'Argument 2 to "Prefetch" has incompatible type "ValuesQuerySet'
    ],
    'proxy_models': [
        'Incompatible types in assignment',
        'in base class "User"'
    ],
    'queries': [
        'Incompatible types in assignment (expression has type "None", variable has type "str")',
        'Invalid index type "Optional[str]" for "Dict[str, int]"; expected type "str"',
        'Unsupported operand types for & ("Manager[Author]" and "Manager[Tag]")',
        'Unsupported operand types for | ("Manager[Author]" and "Manager[Tag]")',
        'ObjectA',
        "'flat' and 'named' can't be used together",
        '"Collection[Any]" has no attribute "explain"',
        "Cannot resolve keyword 'unknown_field' into field",
        'Incompatible type for lookup \'tag\': (got "str", expected "Union[Tag, int, None]")',
        'No overload variant of "__getitem__" of "QuerySet" matches argument type "str"',
    ],
    'requests': [
        'Incompatible types in assignment (expression has type "Dict[str, str]", variable has type "QueryDict")'
    ],
    'responses': [
        'Argument 1 to "TextIOWrapper" has incompatible type "HttpResponse"; expected "IO[bytes]"',
        '"FileLike" has no attribute "closed"',
    ],
    'reverse_lookup': [
        "Cannot resolve keyword 'choice' into field"
    ],
    'settings_tests': [
        'Argument 1 to "Settings" has incompatible type "Optional[str]"; expected "str"'
    ],
    'signals': [
        'Argument 1 to "append" of "list" has incompatible type "Tuple[Any, Any, Optional[Any], Any]";'
    ],
    'sites_framework': [
        'expression has type "CurrentSiteManager[CustomArticle]", base class "AbstractArticle"',
        "Name 'Optional' is not defined",
    ],
    'syndication_tests': [
        'List or tuple expected as variable arguments'
    ],
    'sessions_tests': [
        'Incompatible types in assignment (expression has type "None", variable has type "int")',
        '"AbstractBaseSession" has no attribute',
        '"None" not callable',
    ],
    'select_related': [
        'Item "ForeignKey[Union[Genus, Combinable], Genus]" '
        + 'of "Union[ForeignKey[Union[Genus, Combinable], Genus], Any]"'
    ],
    'select_related_onetoone': [
        'Incompatible types in assignment (expression has type "Parent2", variable has type "Parent1")',
        '"Parent1" has no attribute'
    ],
    'servers': [
        re.compile('Argument [0-9] to "WSGIRequestHandler"'),
        '"HTTPResponse" has no attribute',
        '"type" has no attribute',
        '"WSGIRequest" has no attribute "makefile"',
        'LiveServerAddress',
        '"Stub" has no attribute "makefile"',
    ],
    'serializers': [
        '"Model" has no attribute "data"',
        '"Iterable[Any]" has no attribute "content"',
        re.compile(r'Argument 1 to "(serialize|deserialize)" has incompatible type "None"; expected "str"')
    ],
    'string_lookup': [
        '"Bar" has no attribute "place"',
    ],
    'test_utils': [
        '"PossessedCar" has no attribute "color"',
        'expression has type "None", variable has type "List[str]"',
    ],
    'transactions': [
        'Incompatible types in assignment (expression has type "Thread", variable has type "Callable[[], Any]")'
    ],
    'urlpatterns': [
        '"object" not callable',
        '"None" not callable',
    ],
    'urlpatterns_reverse': [
        'List or tuple expected as variable arguments',
        'No overload variant of "zip" matches argument types "Any", "object"',
        'Argument 1 to "get_callable" has incompatible type "int"'
    ],
    'utils_tests': [
        'Argument 1 to "activate" has incompatible type "None"; expected "Union[tzinfo, str]"',
        'Argument 1 to "activate" has incompatible type "Optional[str]"; expected "str"',
        'Incompatible types in assignment (expression has type "None", base class "object" defined the type as',
        'Class',
        'has no attribute "cp"',
        'Argument "name" to "cached_property" has incompatible type "int"; expected "Optional[str]"',
        'has no attribute "sort"',
        'Unsupported target for indexed assignment',
        'defined the type as "None"',
        'Argument 1 to "Path" has incompatible type "Optional[str]"',
        '"None" not callable',
        '"WSGIRequest" has no attribute "process_response_content"',
        'No overload variant of "join" matches argument types "str", "None"',
        'Argument 1 to "Archive" has incompatible type "None"; expected "str"',
        'Argument 1 to "to_path" has incompatible type "int"; expected "Union[Path, str]"',

    ],
    'view_tests': [
        "Module 'django.views.debug' has no attribute 'Path'",
        'Value of type "Optional[List[str]]" is not indexable',
        'ExceptionUser',
        'view_tests.tests.test_debug.User',
        'Exception must be derived from BaseException',
        "No binding for nonlocal 'tb_frames' found",
    ],
    'validation': [
        'has no attribute "name"',
    ],
    'wsgi': [
        '"HttpResponse" has no attribute "block_size"',
    ],
}


def check_if_custom_ignores_are_covered_by_common() -> None:
    from scripts.typecheck_tests import is_pattern_fits

    common_ignore_patterns = IGNORED_ERRORS['__common__']
    for module_name, patterns in IGNORED_ERRORS.items():
        if module_name == '__common__':
            continue
        for pattern in patterns:
            for common_pattern in common_ignore_patterns:
                if isinstance(pattern, str) and is_pattern_fits(common_pattern, pattern):
                    print(f'pattern "{module_name}: {pattern!r}" is covered by pattern {common_pattern!r}')


check_if_custom_ignores_are_covered_by_common()



from typing import Any

from django.db.backends.base.base import BaseDatabaseWrapper

class BaseDatabaseFeatures:
    gis_enabled: bool = ...
    allows_group_by_pk: bool = ...
    allows_group_by_selected_pks: bool = ...
    empty_fetchmany_value: Any = ...
    update_can_self_select: bool = ...
    interprets_empty_strings_as_nulls: bool = ...
    supports_nullable_unique_constraints: bool = ...
    supports_partially_nullable_unique_constraints: bool = ...
    can_use_chunked_reads: bool = ...
    can_return_id_from_insert: bool = ...
    can_return_ids_from_bulk_insert: bool = ...
    has_bulk_insert: bool = ...
    uses_savepoints: bool = ...
    can_release_savepoints: bool = ...
    related_fields_match_type: bool = ...
    allow_sliced_subqueries_with_in: bool = ...
    has_select_for_update: bool = ...
    has_select_for_update_nowait: bool = ...
    has_select_for_update_skip_locked: bool = ...
    has_select_for_update_of: bool = ...
    select_for_update_of_column: bool = ...
    test_db_allows_multiple_connections: bool = ...
    supports_unspecified_pk: bool = ...
    supports_forward_references: bool = ...
    truncates_names: bool = ...
    has_real_datatype: bool = ...
    supports_subqueries_in_group_by: bool = ...
    has_native_uuid_field: bool = ...
    has_native_duration_field: bool = ...
    supports_temporal_subtraction: bool = ...
    supports_regex_backreferencing: bool = ...
    supports_date_lookup_using_string: bool = ...
    supports_timezones: bool = ...
    has_zoneinfo_database: bool = ...
    requires_explicit_null_ordering_when_grouping: bool = ...
    nulls_order_largest: bool = ...
    max_query_params: Any = ...
    allows_auto_pk_0: bool = ...
    can_defer_constraint_checks: bool = ...
    supports_mixed_date_datetime_comparisons: bool = ...
    supports_tablespaces: bool = ...
    supports_sequence_reset: bool = ...
    can_introspect_default: bool = ...
    can_introspect_foreign_keys: bool = ...
    can_introspect_autofield: bool = ...
    can_introspect_big_integer_field: bool = ...
    can_introspect_binary_field: bool = ...
    can_introspect_decimal_field: bool = ...
    can_introspect_ip_address_field: bool = ...
    can_introspect_positive_integer_field: bool = ...
    can_introspect_small_integer_field: bool = ...
    can_introspect_time_field: bool = ...
    introspected_boolean_field_type: str = ...
    supports_index_column_ordering: bool = ...
    can_distinct_on_fields: bool = ...
    autocommits_when_autocommit_is_off: bool = ...
    atomic_transactions: bool = ...
    can_rollback_ddl: bool = ...
    supports_atomic_references_rename: bool = ...
    supports_combined_alters: bool = ...
    supports_foreign_keys: bool = ...
    supports_column_check_constraints: bool = ...
    supports_paramstyle_pyformat: bool = ...
    requires_literal_defaults: bool = ...
    connection_persists_old_columns: bool = ...
    closed_cursor_error_class: Any = ...
    has_case_insensitive_like: bool = ...
    requires_sqlparse_for_splitting: bool = ...
    bare_select_suffix: str = ...
    implied_column_null: bool = ...
    uppercases_column_names: bool = ...
    supports_select_for_update_with_limit: bool = ...
    greatest_least_ignores_nulls: bool = ...
    can_clone_databases: bool = ...
    ignores_table_name_case: bool = ...
    for_update_after_from: bool = ...
    supports_select_union: bool = ...
    supports_select_intersection: bool = ...
    supports_select_difference: bool = ...
    supports_slicing_ordering_in_compound: bool = ...
    supports_aggregate_filter_clause: bool = ...
    supports_index_on_text_field: bool = ...
    supports_over_clause: bool = ...
    supports_cast_with_precision: bool = ...
    create_test_procedure_without_params_sql: Any = ...
    create_test_procedure_with_int_param_sql: Any = ...
    supports_callproc_kwargs: bool = ...
    db_functions_convert_bytes_to_str: bool = ...
    supported_explain_formats: Any = ...
    validates_explain_options: bool = ...
    connection: Any = ...
    def __init__(self, connection: BaseDatabaseWrapper) -> None: ...
    def supports_explaining_query_execution(self) -> bool: ...
    def supports_transactions(self): ...
    def supports_stddev(self): ...



from typing import Any, Callable, Dict, List, Optional, Sequence, Set, Tuple, Type, TypeVar, Union, Collection

from django.core.checks.messages import CheckMessage
from django.core.exceptions import ValidationError
from django.db.models.manager import Manager
from django.db.models.options import Options

_Self = TypeVar("_Self", bound="Model")

class ModelBase(type): ...

class Model(metaclass=ModelBase):
    class DoesNotExist(Exception): ...
    class MultipleObjectsReturned(Exception): ...
    class Meta: ...
    _default_manager: Manager[Model]
    _meta: Options[Any]
    objects: Manager[Any]
    pk: Any = ...
    def __init__(self: _Self, *args, **kwargs) -> None: ...
    def delete(self, using: Any = ..., keep_parents: bool = ...) -> Tuple[int, Dict[str, int]]: ...
    def full_clean(self, exclude: Optional[Collection[str]] = ..., validate_unique: bool = ...) -> None: ...
    def clean(self) -> None: ...
    def clean_fields(self, exclude: Optional[Collection[str]] = ...) -> None: ...
    def validate_unique(self, exclude: Optional[Collection[str]] = ...) -> None: ...
    def unique_error_message(
        self, model_class: Type[_Self], unique_check: Collection[Union[Callable, str]]
    ) -> ValidationError: ...
    def save(
        self,
        force_insert: bool = ...,
        force_update: bool = ...,
        using: Optional[str] = ...,
        update_fields: Optional[Union[Sequence[str], str]] = ...,
    ) -> None: ...
    def save_base(
        self,
        raw: bool = ...,
        force_insert: bool = ...,
        force_update: bool = ...,
        using: Optional[str] = ...,
        update_fields: Optional[Union[Sequence[str], str]] = ...,
    ): ...
    def refresh_from_db(self: _Self, using: Optional[str] = ..., fields: Optional[List[str]] = ...) -> _Self: ...
    def get_deferred_fields(self) -> Set[str]: ...
    @classmethod
    def check(cls, **kwargs: Any) -> List[CheckMessage]: ...
    def __getstate__(self) -> dict: ...

class ModelStateFieldsCacheDescriptor: ...

class ModelState:
    db: None = ...
    adding: bool = ...
    fields_cache: ModelStateFieldsCacheDescriptor = ...



from io import BytesIO
from typing import Any, Dict, List, Optional, Pattern, Tuple, Type, Union

from django.contrib.auth.models import AbstractUser
from django.contrib.sessions.backends.base import SessionBase
from django.core.handlers.base import BaseHandler
from django.core.serializers.json import DjangoJSONEncoder
from django.http.cookie import SimpleCookie
from django.http.request import HttpRequest
from django.http.response import HttpResponse, HttpResponseBase

from django.core.handlers.wsgi import WSGIRequest

BOUNDARY: str = ...
MULTIPART_CONTENT: str = ...
CONTENT_TYPE_RE: Pattern = ...
JSON_CONTENT_TYPE_RE: Pattern = ...

class RedirectCycleError(Exception):
    last_response: HttpResponseBase = ...
    redirect_chain: List[Tuple[str, int]] = ...
    def __init__(self, message: str, last_response: HttpResponseBase) -> None: ...

class FakePayload:
    read_started: bool = ...
    def __init__(self, content: Optional[Union[bytes, str]] = ...) -> None: ...
    def __len__(self) -> int: ...
    def read(self, num_bytes: int = ...) -> bytes: ...
    def write(self, content: Union[bytes, str]) -> None: ...

class ClientHandler(BaseHandler):
    enforce_csrf_checks: bool = ...
    def __init__(self, enforce_csrf_checks: bool = ..., *args: Any, **kwargs: Any) -> None: ...
    def __call__(self, environ: Dict[str, Any]) -> HttpResponseBase: ...

def encode_multipart(boundary: str, data: Dict[str, Any]) -> bytes: ...
def encode_file(boundary: str, key: str, file: Any) -> List[bytes]: ...

class RequestFactory:
    json_encoder: Type[DjangoJSONEncoder] = ...
    defaults: Dict[str, str] = ...
    cookies: SimpleCookie = ...
    errors: BytesIO = ...
    def __init__(self, *, json_encoder: Any = ..., **defaults: Any) -> None: ...
    def request(self, **request: Any) -> WSGIRequest: ...
    def get(self, path: str, data: Any = ..., secure: bool = ..., **extra: Any) -> WSGIRequest: ...
    def post(
        self, path: str, data: Any = ..., content_type: str = ..., secure: bool = ..., **extra: Any
    ) -> WSGIRequest: ...
    def head(self, path: str, data: Any = ..., secure: bool = ..., **extra: Any) -> WSGIRequest: ...
    def trace(self, path: str, secure: bool = ..., **extra: Any) -> WSGIRequest: ...
    def options(
        self,
        path: str,
        data: Union[Dict[str, str], str] = ...,
        content_type: str = ...,
        secure: bool = ...,
        **extra: Any
    ) -> WSGIRequest: ...
    def put(
        self, path: str, data: Any = ..., content_type: str = ..., secure: bool = ..., **extra: Any
    ) -> WSGIRequest: ...
    def patch(
        self, path: str, data: Any = ..., content_type: str = ..., secure: bool = ..., **extra: Any
    ) -> WSGIRequest: ...
    def delete(
        self, path: str, data: Any = ..., content_type: str = ..., secure: bool = ..., **extra: Any
    ) -> WSGIRequest: ...
    def generic(
        self,
        method: str,
        path: str,
        data: Any = ...,
        content_type: Optional[str] = ...,
        secure: bool = ...,
        **extra: Any
    ) -> WSGIRequest: ...

class Client:
    json_encoder: Type[DjangoJSONEncoder] = ...
    defaults: Dict[str, str] = ...
    cookies: SimpleCookie = ...
    errors: BytesIO = ...
    handler: ClientHandler = ...
    exc_info: None = ...
    def __init__(self, enforce_csrf_checks: bool = ..., **defaults: Any) -> None: ...
    def request(self, **request: Any) -> Any: ...
    def get(self, path: str, data: Any = ..., secure: bool = ..., **extra: Any) -> Any: ...
    def post(self, path: str, data: Any = ..., content_type: str = ..., secure: bool = ..., **extra: Any) -> Any: ...
    def head(self, path: str, data: Any = ..., secure: bool = ..., **extra: Any) -> Any: ...
    def trace(self, path: str, secure: bool = ..., **extra: Any) -> Any: ...
    def options(
        self,
        path: str,
        data: Union[Dict[str, str], str] = ...,
        content_type: str = ...,
        secure: bool = ...,
        **extra: Any
    ) -> Any: ...
    def put(self, path: str, data: Any = ..., content_type: str = ..., secure: bool = ..., **extra: Any) -> Any: ...
    def patch(self, path: str, data: Any = ..., content_type: str = ..., secure: bool = ..., **extra: Any) -> Any: ...
    def delete(self, path: str, data: Any = ..., content_type: str = ..., secure: bool = ..., **extra: Any) -> Any: ...
    def generic(
        self,
        method: str,
        path: str,
        data: Any = ...,
        content_type: Optional[str] = ...,
        secure: bool = ...,
        **extra: Any
    ) -> Any: ...
    def store_exc_info(self, **kwargs: Any) -> None: ...
    @property
    def session(self) -> SessionBase: ...
    def login(self, **credentials: Any) -> bool: ...
    def force_login(self, user: AbstractUser, backend: Optional[str] = ...) -> None: ...
    def logout(self) -> None: ...

def conditional_content_removal(request: HttpRequest, response: HttpResponseBase) -> HttpResponse: ...



import os
from distutils.core import setup
from typing import List

from setuptools import find_packages


def find_stub_files(name: str) -> List[str]:
    result = []
    for root, dirs, files in os.walk(name):
        for file in files:
            if file.endswith('.pyi'):
                if os.path.sep in root:
                    sub_root = root.split(os.path.sep, 1)[-1]
                    file = os.path.join(sub_root, file)
                result.append(file)
    return result


with open('README.md', 'r') as f:
    readme = f.read()

dependencies = [
    'mypy>=0.750,<0.760',
    'typing-extensions',
    'django',
]

setup(
    name="django-stubs",
    version="1.3.0",
    description='Mypy stubs for Django',
    long_description=readme,
    long_description_content_type='text/markdown',
    license='MIT',
    url="https://github.com/typeddjango/django-stubs",
    author="Maksim Kurnikov",
    author_email="maxim.kurnikov@gmail.com",
    py_modules=[],
    python_requires='>=3.6',
    install_requires=dependencies,
    packages=['django-stubs', *find_packages(exclude=['scripts'])],
    package_data={'django-stubs': find_stub_files('django-stubs')},
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ]
)



from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Iterable, Union

from django.db.models.base import Model
from django.db.models.query import QuerySet

_T = TypeVar("_T", bound=Model, covariant=True)
_M = TypeVar("_M", bound="BaseManager")

class BaseManager(QuerySet[_T]):
    creation_counter: int = ...
    auto_created: bool = ...
    use_in_migrations: bool = ...
    name: str = ...
    model: Type[_T] = ...
    db: str
    def __init__(self) -> None: ...
    def deconstruct(self) -> Tuple[bool, str, None, Tuple, Dict[str, int]]: ...
    def check(self, **kwargs: Any) -> List[Any]: ...
    @classmethod
    def from_queryset(cls, queryset_class: Type[QuerySet], class_name: Optional[str] = ...) -> Any: ...
    @classmethod
    def _get_queryset_methods(cls, queryset_class: type) -> Dict[str, Any]: ...
    def contribute_to_class(self, model: Type[Model], name: str) -> None: ...
    def db_manager(self: _M, using: Optional[str] = ..., hints: Optional[Dict[str, Model]] = ...) -> _M: ...
    def get_queryset(self) -> QuerySet[_T]: ...

class Manager(BaseManager[_T]): ...

class RelatedManager(Manager[_T]):
    related_val: Tuple[int, ...]
    def add(self, *objs: Union[_T, int], bulk: bool = ...) -> None: ...
    def remove(self, *objs: Union[_T, int], bulk: bool = ...) -> None: ...
    def set(
        self, objs: Union[QuerySet[_T], Iterable[Union[_T, int]]], *, bulk: bool = ..., clear: bool = ...
    ) -> None: ...
    def clear(self) -> None: ...

class ManagerDescriptor:
    manager: Manager = ...
    def __init__(self, manager: Manager) -> None: ...
    def __get__(self, instance: Optional[Model], cls: Type[Model] = ...) -> Manager: ...

class EmptyManager(Manager):
    def __init__(self, model: Type[Model]) -> None: ...



from typing import Dict, List, Optional, Protocol, Sequence, Union

from django.db.models.base import Model
from django.db.models.query import QuerySet

class UnorderedObjectListWarning(RuntimeWarning): ...
class InvalidPage(Exception): ...
class PageNotAnInteger(InvalidPage): ...
class EmptyPage(InvalidPage): ...

class SupportsLen(Protocol):
    def __len__(self) -> int: ...

class SupportsCount(Protocol):
    def count(self) -> int: ...

class SupportsOrdered(Protocol):
    ordered: bool = ...

class Paginator:
    object_list: QuerySet = ...
    per_page: int = ...
    orphans: int = ...
    allow_empty_first_page: bool = ...
    def __init__(
        self,
        object_list: Union[SupportsLen, SupportsCount, SupportsOrdered],
        per_page: Union[int, str],
        orphans: int = ...,
        allow_empty_first_page: bool = ...,
    ) -> None: ...
    def validate_number(self, number: Optional[Union[float, str]]) -> int: ...
    def get_page(self, number: Optional[int]) -> Page: ...
    def page(self, number: Union[int, str]) -> Page: ...
    @property
    def count(self) -> int: ...
    @property
    def num_pages(self) -> int: ...
    @property
    def page_range(self) -> range: ...

QuerySetPaginator = Paginator

class Page(Sequence):
    object_list: QuerySet = ...
    number: int = ...
    paginator: Paginator = ...
    def __init__(
        self,
        object_list: Union[List[Dict[str, str]], List[Model], List[int], QuerySet, str],
        number: int,
        paginator: Paginator,
    ) -> None: ...
    def __getitem__(self, item): ...
    def __len__(self): ...
    def has_next(self) -> bool: ...
    def has_previous(self) -> bool: ...
    def has_other_pages(self) -> bool: ...
    def next_page_number(self) -> int: ...
    def previous_page_number(self) -> int: ...
    def start_index(self) -> int: ...
    def end_index(self) -> int: ...



from typing import Any, Optional, Set, Union

from django.contrib.auth.base_user import AbstractBaseUser
from django.contrib.auth.models import AnonymousUser, User

from django.db.models.base import Model

_AnyUser = Union[Model, AnonymousUser]

UserModel: Any

class ModelBackend:
    def authenticate(
        self, request: Any, username: Optional[str] = ..., password: Optional[str] = ..., **kwargs: Any
    ) -> Optional[AbstractBaseUser]: ...
    def user_can_authenticate(self, user: Optional[_AnyUser]) -> bool: ...
    def get_user_permissions(self, user_obj: _AnyUser, obj: Optional[Model] = ...) -> Set[str]: ...
    def get_group_permissions(self, user_obj: _AnyUser, obj: Optional[Model] = ...) -> Set[str]: ...
    def get_all_permissions(self, user_obj: _AnyUser, obj: Optional[Model] = ...) -> Set[str]: ...
    def has_perm(self, user_obj: _AnyUser, perm: str, obj: Optional[Model] = ...) -> bool: ...
    def has_module_perms(self, user_obj: _AnyUser, app_label: str) -> bool: ...
    def get_user(self, user_id: int) -> AbstractBaseUser: ...

class AllowAllUsersModelBackend(ModelBackend): ...

class RemoteUserBackend(ModelBackend):
    create_unknown_user: bool = ...
    def clean_username(self, username: str) -> str: ...
    def configure_user(self, user: User) -> User: ...

class AllowAllUsersRemoteUserBackend(RemoteUserBackend): ...



import os
import sys
from collections import defaultdict
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING, Dict, Iterable, Iterator, Optional, Set, Tuple, Type, Union,
)

from django.core.exceptions import FieldError
from django.db import models
from django.db.models.base import Model
from django.db.models.fields import AutoField, CharField, Field
from django.db.models.fields.related import ForeignKey, RelatedField
from django.db.models.fields.reverse_related import ForeignObjectRel
from django.db.models.lookups import Exact
from django.db.models.sql.query import Query
from django.utils.functional import cached_property
from mypy.checker import TypeChecker
from mypy.plugin import MethodContext
from mypy.types import AnyType, Instance
from mypy.types import Type as MypyType
from mypy.types import TypeOfAny, UnionType

from mypy_django_plugin.lib import fullnames, helpers

try:
    from django.contrib.postgres.fields import ArrayField
except ImportError:
    class ArrayField:  # type: ignore
        pass

if TYPE_CHECKING:
    from django.apps.registry import Apps  # noqa: F401
    from django.conf import LazySettings  # noqa: F401


@contextmanager
def temp_environ():
    """Allow the ability to set os.environ temporarily"""
    environ = dict(os.environ)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(environ)


def initialize_django(settings_module: str) -> Tuple['Apps', 'LazySettings']:
    with temp_environ():
        os.environ['DJANGO_SETTINGS_MODULE'] = settings_module

        # add current directory to sys.path
        sys.path.append(os.getcwd())

        def noop_class_getitem(cls, key):
            return cls

        from django.db import models

        models.QuerySet.__class_getitem__ = classmethod(noop_class_getitem)  # type: ignore
        models.Manager.__class_getitem__ = classmethod(noop_class_getitem)  # type: ignore

        from django.conf import settings
        from django.apps import apps

        apps.get_models.cache_clear()  # type: ignore
        apps.get_swappable_settings_name.cache_clear()  # type: ignore

        if not settings.configured:
            settings._setup()

        apps.populate(settings.INSTALLED_APPS)

    assert apps.apps_ready
    assert settings.configured

    return apps, settings


class LookupsAreUnsupported(Exception):
    pass


class DjangoContext:
    def __init__(self, django_settings_module: str) -> None:
        self.django_settings_module = django_settings_module

        apps, settings = initialize_django(self.django_settings_module)
        self.apps_registry = apps
        self.settings = settings

    @cached_property
    def model_modules(self) -> Dict[str, Set[Type[Model]]]:
        """ All modules that contain Django models. """
        if self.apps_registry is None:
            return {}

        modules: Dict[str, Set[Type[Model]]] = defaultdict(set)
        for concrete_model_cls in self.apps_registry.get_models():
            modules[concrete_model_cls.__module__].add(concrete_model_cls)
            # collect abstract=True models
            for model_cls in concrete_model_cls.mro()[1:]:
                if (issubclass(model_cls, Model)
                        and hasattr(model_cls, '_meta')
                        and model_cls._meta.abstract):
                    modules[model_cls.__module__].add(model_cls)
        return modules

    def get_model_class_by_fullname(self, fullname: str) -> Optional[Type[Model]]:
        # Returns None if Model is abstract
        module, _, model_cls_name = fullname.rpartition('.')
        for model_cls in self.model_modules.get(module, set()):
            if model_cls.__name__ == model_cls_name:
                return model_cls
        return None

    def get_model_fields(self, model_cls: Type[Model]) -> Iterator[Field]:
        for field in model_cls._meta.get_fields():
            if isinstance(field, Field):
                yield field

    def get_model_relations(self, model_cls: Type[Model]) -> Iterator[ForeignObjectRel]:
        for field in model_cls._meta.get_fields():
            if isinstance(field, ForeignObjectRel):
                yield field

    def get_field_lookup_exact_type(self, api: TypeChecker, field: Field) -> MypyType:
        if isinstance(field, (RelatedField, ForeignObjectRel)):
            related_model_cls = field.related_model
            primary_key_field = self.get_primary_key_field(related_model_cls)
            primary_key_type = self.get_field_get_type(api, primary_key_field, method='init')

            rel_model_info = helpers.lookup_class_typeinfo(api, related_model_cls)
            if rel_model_info is None:
                return AnyType(TypeOfAny.explicit)

            model_and_primary_key_type = UnionType.make_union([Instance(rel_model_info, []),
                                                               primary_key_type])
            return helpers.make_optional(model_and_primary_key_type)
            # return helpers.make_optional(Instance(rel_model_info, []))

        field_info = helpers.lookup_class_typeinfo(api, field.__class__)
        if field_info is None:
            return AnyType(TypeOfAny.explicit)
        return helpers.get_private_descriptor_type(field_info, '_pyi_lookup_exact_type',
                                                   is_nullable=field.null)

    def get_primary_key_field(self, model_cls: Type[Model]) -> Field:
        for field in model_cls._meta.get_fields():
            if isinstance(field, Field):
                if field.primary_key:
                    return field
        raise ValueError('No primary key defined')

    def get_expected_types(self, api: TypeChecker, model_cls: Type[Model], *, method: str) -> Dict[str, MypyType]:
        from django.contrib.contenttypes.fields import GenericForeignKey

        expected_types = {}
        # add pk if not abstract=True
        if not model_cls._meta.abstract:
            primary_key_field = self.get_primary_key_field(model_cls)
            field_set_type = self.get_field_set_type(api, primary_key_field, method=method)
            expected_types['pk'] = field_set_type

        for field in model_cls._meta.get_fields():
            if isinstance(field, Field):
                field_name = field.attname
                field_set_type = self.get_field_set_type(api, field, method=method)
                expected_types[field_name] = field_set_type

                if isinstance(field, ForeignKey):
                    field_name = field.name
                    foreign_key_info = helpers.lookup_class_typeinfo(api, field.__class__)
                    if foreign_key_info is None:
                        # maybe there's no type annotation for the field
                        expected_types[field_name] = AnyType(TypeOfAny.unannotated)
                        continue

                    related_model = self.get_field_related_model_cls(field)
                    if related_model is None:
                        expected_types[field_name] = AnyType(TypeOfAny.from_error)
                        continue

                    if related_model._meta.proxy_for_model is not None:
                        related_model = related_model._meta.proxy_for_model

                    related_model_info = helpers.lookup_class_typeinfo(api, related_model)
                    if related_model_info is None:
                        expected_types[field_name] = AnyType(TypeOfAny.unannotated)
                        continue

                    is_nullable = self.get_field_nullability(field, method)
                    foreign_key_set_type = helpers.get_private_descriptor_type(foreign_key_info,
                                                                               '_pyi_private_set_type',
                                                                               is_nullable=is_nullable)
                    model_set_type = helpers.convert_any_to_type(foreign_key_set_type,
                                                                 Instance(related_model_info, []))

                    expected_types[field_name] = model_set_type

            elif isinstance(field, GenericForeignKey):
                # it's generic, so cannot set specific model
                field_name = field.name
                gfk_info = helpers.lookup_class_typeinfo(api, field.__class__)
                gfk_set_type = helpers.get_private_descriptor_type(gfk_info, '_pyi_private_set_type',
                                                                   is_nullable=True)
                expected_types[field_name] = gfk_set_type

        return expected_types

    @cached_property
    def all_registered_model_classes(self) -> Set[Type[models.Model]]:
        model_classes = self.apps_registry.get_models()

        all_model_bases = set()
        for model_cls in model_classes:
            for base_cls in model_cls.mro():
                if issubclass(base_cls, models.Model):
                    all_model_bases.add(base_cls)

        return all_model_bases

    @cached_property
    def all_registered_model_class_fullnames(self) -> Set[str]:
        return {helpers.get_class_fullname(cls) for cls in self.all_registered_model_classes}

    def get_attname(self, field: Field) -> str:
        attname = field.attname
        return attname

    def get_field_nullability(self, field: Field, method: Optional[str]) -> bool:
        nullable = field.null
        if not nullable and isinstance(field, CharField) and field.blank:
            return True
        if method == '__init__':
            if field.primary_key or isinstance(field, ForeignKey):
                return True
        if method == 'create':
            if isinstance(field, AutoField):
                return True
        if field.has_default():
            return True
        return nullable

    def get_field_set_type(self, api: TypeChecker, field: Field, *, method: str) -> MypyType:
        """ Get a type of __set__ for this specific Django field. """
        target_field = field
        if isinstance(field, ForeignKey):
            target_field = field.target_field

        field_info = helpers.lookup_class_typeinfo(api, target_field.__class__)
        if field_info is None:
            return AnyType(TypeOfAny.from_error)

        field_set_type = helpers.get_private_descriptor_type(field_info, '_pyi_private_set_type',
                                                             is_nullable=self.get_field_nullability(field, method))
        if isinstance(target_field, ArrayField):
            argument_field_type = self.get_field_set_type(api, target_field.base_field, method=method)
            field_set_type = helpers.convert_any_to_type(field_set_type, argument_field_type)
        return field_set_type

    def get_field_get_type(self, api: TypeChecker, field: Field, *, method: str) -> MypyType:
        """ Get a type of __get__ for this specific Django field. """
        field_info = helpers.lookup_class_typeinfo(api, field.__class__)
        if field_info is None:
            return AnyType(TypeOfAny.unannotated)

        is_nullable = self.get_field_nullability(field, method)
        if isinstance(field, RelatedField):
            related_model_cls = self.get_field_related_model_cls(field)
            if related_model_cls is None:
                return AnyType(TypeOfAny.from_error)

            if method == 'values':
                primary_key_field = self.get_primary_key_field(related_model_cls)
                return self.get_field_get_type(api, primary_key_field, method=method)

            model_info = helpers.lookup_class_typeinfo(api, related_model_cls)
            if model_info is None:
                return AnyType(TypeOfAny.unannotated)

            return Instance(model_info, [])
        else:
            return helpers.get_private_descriptor_type(field_info, '_pyi_private_get_type',
                                                       is_nullable=is_nullable)

    def get_field_related_model_cls(self, field: Union[RelatedField, ForeignObjectRel]) -> Optional[Type[Model]]:
        if isinstance(field, RelatedField):
            related_model_cls = field.remote_field.model
        else:
            related_model_cls = field.field.model

        if isinstance(related_model_cls, str):
            if related_model_cls == 'self':
                # same model
                related_model_cls = field.model
            elif '.' not in related_model_cls:
                # same file model
                related_model_fullname = field.model.__module__ + '.' + related_model_cls
                related_model_cls = self.get_model_class_by_fullname(related_model_fullname)
            else:
                related_model_cls = self.apps_registry.get_model(related_model_cls)

        return related_model_cls

    def _resolve_field_from_parts(self, field_parts: Iterable[str], model_cls: Type[Model]) -> Field:
        currently_observed_model = model_cls
        field = None
        for field_part in field_parts:
            if field_part == 'pk':
                field = self.get_primary_key_field(currently_observed_model)
                continue

            field = currently_observed_model._meta.get_field(field_part)
            if isinstance(field, RelatedField):
                currently_observed_model = field.related_model
                model_name = currently_observed_model._meta.model_name
                if (model_name is not None
                        and field_part == (model_name + '_id')):
                    field = self.get_primary_key_field(currently_observed_model)

            if isinstance(field, ForeignObjectRel):
                currently_observed_model = field.related_model

        assert field is not None
        return field

    def resolve_lookup_into_field(self, model_cls: Type[Model], lookup: str) -> Field:
        query = Query(model_cls)
        lookup_parts, field_parts, is_expression = query.solve_lookup_type(lookup)
        if lookup_parts:
            raise LookupsAreUnsupported()

        return self._resolve_field_from_parts(field_parts, model_cls)

    def resolve_lookup_expected_type(self, ctx: MethodContext, model_cls: Type[Model], lookup: str) -> MypyType:
        query = Query(model_cls)
        try:
            lookup_parts, field_parts, is_expression = query.solve_lookup_type(lookup)
            if is_expression:
                return AnyType(TypeOfAny.explicit)
        except FieldError as exc:
            ctx.api.fail(exc.args[0], ctx.context)
            return AnyType(TypeOfAny.from_error)

        field = self._resolve_field_from_parts(field_parts, model_cls)

        lookup_cls = None
        if lookup_parts:
            lookup = lookup_parts[-1]
            lookup_cls = field.get_lookup(lookup)
            if lookup_cls is None:
                # unknown lookup
                return AnyType(TypeOfAny.explicit)

        if lookup_cls is None or isinstance(lookup_cls, Exact):
            return self.get_field_lookup_exact_type(helpers.get_typechecker_api(ctx), field)

        assert lookup_cls is not None

        lookup_info = helpers.lookup_class_typeinfo(helpers.get_typechecker_api(ctx), lookup_cls)
        if lookup_info is None:
            return AnyType(TypeOfAny.explicit)

        for lookup_base in helpers.iter_bases(lookup_info):
            if lookup_base.args and isinstance(lookup_base.args[0], Instance):
                lookup_type: MypyType = lookup_base.args[0]
                # if it's Field, consider lookup_type a __get__ of current field
                if (isinstance(lookup_type, Instance)
                        and lookup_type.type.fullname == fullnames.FIELD_FULLNAME):
                    field_info = helpers.lookup_class_typeinfo(helpers.get_typechecker_api(ctx), field.__class__)
                    if field_info is None:
                        return AnyType(TypeOfAny.explicit)
                    lookup_type = helpers.get_private_descriptor_type(field_info, '_pyi_private_get_type',
                                                                      is_nullable=field.null)
                return lookup_type

        return AnyType(TypeOfAny.explicit)

    def resolve_f_expression_type(self, f_expression_type: Instance) -> MypyType:
        return AnyType(TypeOfAny.explicit)



from io import BytesIO
from typing import (
    Any,
    BinaryIO,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Pattern,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from django.contrib.auth.base_user import AbstractBaseUser
from django.contrib.sessions.backends.base import SessionBase
from django.utils.datastructures import CaseInsensitiveMapping, ImmutableList, MultiValueDict

from django.core.files import uploadedfile, uploadhandler
from django.urls import ResolverMatch

RAISE_ERROR: object = ...
host_validation_re: Pattern = ...

class UnreadablePostError(IOError): ...
class RawPostDataException(Exception): ...

UploadHandlerList = Union[List[uploadhandler.FileUploadHandler], ImmutableList[uploadhandler.FileUploadHandler]]

class HttpHeaders(CaseInsensitiveMapping):
    HTTP_PREFIX: str = ...
    UNPREFIXED_HEADERS: Set[str] = ...
    def __init__(self, environ: Mapping[str, Any]) -> None: ...
    @classmethod
    def parse_header_name(cls, header: str) -> Optional[str]: ...

class HttpRequest(BytesIO):
    GET: QueryDict = ...
    POST: QueryDict = ...
    COOKIES: Dict[str, str] = ...
    META: Dict[str, Any] = ...
    FILES: MultiValueDict[str, uploadedfile.UploadedFile] = ...
    path: str = ...
    path_info: str = ...
    method: Optional[str] = ...
    resolver_match: ResolverMatch = ...
    content_type: Optional[str] = ...
    content_params: Optional[Dict[str, str]] = ...
    user: AbstractBaseUser
    session: SessionBase
    encoding: Optional[str] = ...
    upload_handlers: UploadHandlerList = ...
    def __init__(self) -> None: ...
    def get_host(self) -> str: ...
    def get_port(self) -> str: ...
    def get_full_path(self, force_append_slash: bool = ...) -> str: ...
    def get_full_path_info(self, force_append_slash: bool = ...) -> str: ...
    def get_signed_cookie(
        self, key: str, default: Any = ..., salt: str = ..., max_age: Optional[int] = ...
    ) -> Optional[str]: ...
    def get_raw_uri(self) -> str: ...
    def build_absolute_uri(self, location: Optional[str] = ...) -> str: ...
    @property
    def scheme(self) -> Optional[str]: ...
    def is_secure(self) -> bool: ...
    def is_ajax(self) -> bool: ...
    def parse_file_upload(
        self, META: Mapping[str, Any], post_data: BinaryIO
    ) -> Tuple[QueryDict, MultiValueDict[str, uploadedfile.UploadedFile]]: ...
    @property
    def headers(self) -> HttpHeaders: ...
    @property
    def body(self) -> bytes: ...
    def _load_post_and_files(self) -> None: ...

_Q = TypeVar("_Q", bound="QueryDict")

class QueryDict(MultiValueDict[str, str]):
    encoding: str = ...
    _mutable: bool = ...
    def __init__(
        self, query_string: Optional[Union[str, bytes]] = ..., mutable: bool = ..., encoding: Optional[str] = ...
    ) -> None: ...
    def setlist(self, key: str, list_: List[str]) -> None: ...
    def setlistdefault(self, key: str, default_list: Optional[List[str]] = ...) -> List[str]: ...
    def appendlist(self, key: str, value: str) -> None: ...
    def urlencode(self, safe: Optional[str] = ...) -> str: ...
    @classmethod
    def fromkeys(
        cls: Type[_Q],
        iterable: Iterable[Union[bytes, str]],
        value: Any = ...,
        mutable: bool = ...,
        encoding: Optional[str] = ...,
    ) -> _Q: ...

@overload
def bytes_to_text(s: bytes, encoding: str) -> str: ...
@overload
def bytes_to_text(s: str, encoding: str) -> str: ...
@overload
def bytes_to_text(s: None, encoding: str) -> None: ...
def split_domain_port(host: str) -> Tuple[str, str]: ...
def validate_host(host: str, allowed_hosts: Iterable[str]) -> bool: ...
